/*
    This file is part of CrabEmu.

    Copyright (C) 2009, 2013 Lawrence Sebald

    CrabEmu is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2 
    as published by the Free Software Foundation.

    CrabEmu is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CrabEmu; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

/* Imported from Dream Photo Manager 1.0.0 */

#include <string.h>

#include "pvrutils.h"
#include "font.h"

void draw_colored_quad(uint32 list, float x, float y, float z, float w, float h,
                       uint32 color) {
    pvr_poly_cxt_t cxt;
    pvr_poly_hdr_t hdr;
    pvr_vertex_t vert;

    pvr_poly_cxt_col(&cxt, list);
    pvr_poly_compile(&hdr, &cxt);

    pvr_prim(&hdr, sizeof(pvr_poly_hdr_t));

    vert.flags = PVR_CMD_VERTEX;
    vert.oargb = 0;
    vert.argb = color;
    vert.u = 0.0f;
    vert.v = 0.0f;
    vert.z = z;

    vert.x = x;
    vert.y = y + h;
    pvr_prim(&vert, sizeof(pvr_vertex_t));

    vert.y = y;
    pvr_prim(&vert, sizeof(pvr_vertex_t));

    vert.x = x + w;
    vert.y = y + h;
    pvr_prim(&vert, sizeof(pvr_vertex_t));

    vert.flags = PVR_CMD_VERTEX_EOL;
    vert.y = y;
    pvr_prim(&vert, sizeof(pvr_vertex_t));
}

void draw_multicolored_quad(uint32 list, float x, float y, float z, float w,
                            float h, uint32 c1, uint32 c2, uint32 c3,
                            uint32 c4) {
    pvr_poly_cxt_t cxt;
    pvr_poly_hdr_t hdr;
    pvr_vertex_t vert;
    
    pvr_poly_cxt_col(&cxt, list);
    pvr_poly_compile(&hdr, &cxt);
    
    pvr_prim(&hdr, sizeof(pvr_poly_hdr_t));
    
    vert.flags = PVR_CMD_VERTEX;
    vert.oargb = 0;
    vert.argb = c1;
    vert.u = 0.0f;
    vert.v = 0.0f;
    vert.z = z;
    
    vert.x = x;
    vert.y = y + h;
    pvr_prim(&vert, sizeof(pvr_vertex_t));
    
    vert.y = y;
    vert.argb = c2;
    pvr_prim(&vert, sizeof(pvr_vertex_t));
    
    vert.x = x + w;
    vert.y = y + h;
    vert.argb = c3;
    pvr_prim(&vert, sizeof(pvr_vertex_t));
    
    vert.flags = PVR_CMD_VERTEX_EOL;
    vert.y = y;
    vert.argb = c4;
    pvr_prim(&vert, sizeof(pvr_vertex_t));
}

static void string_stats(const char *str, int *longest, int *count) {
    int i, cur, len = strlen(str);

    *longest = 0;
    *count = 0;
    cur = 0;

    for(i = 0; i < len; ++i) {
        if(str[i] == '\n') {
            if(*longest < cur)
                *longest = cur;
            cur = 0;
            ++(*count);
        }
        else {
            ++cur;
        }
    }

    /* Handle the last line. */
    if(*longest < cur)
        *longest = cur;

    ++(*count);
}

void draw_dialog_centered(const char *str, float xmin, float xmax, float ymin,
                          float ymax, float z, uint32 color, pvr_ptr_t f) {
    int longest, count;
    float w, h;
    float xs, ys;

    string_stats(str, &longest, &count);

    w = longest * FONT_CHAR_WIDTH;
    h = count * FONT_LINE_HEIGHT;

    xs = ((xmin + xmax) / 2.0f) - (w / 2.0f);
    ys = ((ymin + ymax) / 2.0f) - (h / 2.0f);

    draw_colored_quad(PVR_LIST_TR_POLY, xs, ys, z, w, h, color);

    font_draw_str_centered(str, f, (int)xmin, (int)xmax, (int)ys, (int)z + 1);
}
