/*
    This file is part of CrabEmu.

    Copyright (C) 2005, 2006, 2007, 2008, 2009, 2012 Lawrence Sebald

    CrabEmu is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2 
    as published by the Free Software Foundation.

    CrabEmu is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CrabEmu; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#ifndef CRABEMUPREFSCONTROLLER_H
#define CRABEMUPREFSCONTROLLER_H

#import <Cocoa/Cocoa.h>

#include "CrabEmu.h"
#include "gamepad.h"

@class EmuController;

@interface CrabEmuPrefsController : NSObject
{
    IBOutlet NSPanel *prefWindow;

    IBOutlet NSMenu *p1input;
    IBOutlet NSMenu *p2input;

    IBOutlet NSPopUpButton *pjoysticks;
    IBOutlet NSButton *pjoyUp;
    IBOutlet NSButton *pjoyRight;
    IBOutlet NSButton *pjoyDown;
    IBOutlet NSButton *pjoyLeft;
    IBOutlet NSButton *pjoyB1;
    IBOutlet NSButton *pjoyB2;
    IBOutlet NSButton *pjoyGGStart;

    IBOutlet NSButton *pNESPadUp;
    IBOutlet NSButton *pNESPadDown;
    IBOutlet NSButton *pNESPadLeft;
    IBOutlet NSButton *pNESPadRight;
    IBOutlet NSButton *pNESPadStart;
    IBOutlet NSButton *pNESPadSelect;
    IBOutlet NSButton *pNESPadA;
    IBOutlet NSButton *pNESPadB;

    IBOutlet NSButton *pCVPadUp;
    IBOutlet NSButton *pCVPadDown;
    IBOutlet NSButton *pCVPadLeft;
    IBOutlet NSButton *pCVPadRight;
    IBOutlet NSButton *pCVPadLTrig;
    IBOutlet NSButton *pCVPadRTrig;
    IBOutlet NSButton *pCVPad1;
    IBOutlet NSButton *pCVPad2;
    IBOutlet NSButton *pCVPad3;
    IBOutlet NSButton *pCVPad4;
    IBOutlet NSButton *pCVPad5;
    IBOutlet NSButton *pCVPad6;
    IBOutlet NSButton *pCVPad7;
    IBOutlet NSButton *pCVPad8;
    IBOutlet NSButton *pCVPad9;
    IBOutlet NSButton *pCVPadStar;
    IBOutlet NSButton *pCVPad0;
    IBOutlet NSButton *pCVPadPound;

    IBOutlet NSButton *keyboardWithGamepad;

    IBOutlet NSPanel *joyPanel;

    IBOutlet NSPanel *buttonAssignment;
    IBOutlet NSTextField *buttonBox;

    IBOutlet NSTextField *smsBios;
    IBOutlet NSButton *smsBiosEnabled;
    IBOutlet NSTextField *ggBios;
    IBOutlet NSButton *ggBiosEnabled;
    IBOutlet NSTextField *cvBios;
    IBOutlet NSTextField *saveDir;
    IBOutlet NSTextField *saveStateDir;
    IBOutlet NSTextField *cheatDir;
    IBOutlet NSButton *ssaverDisabled;
    IBOutlet NSButton *startFullscreen;
    IBOutlet NSButton *preserveAspect;

    IBOutlet EmuController *mainController;

    NSUserDefaults *prefs;
    int pad_count;

    NSMutableDictionary *keymap;
    NSMutableDictionary *cvmap;
    NSMutableDictionary *nesmap;

    unichar smsButtons[2][9];
    unichar cvButtons[2][18];
    unichar nesButtons[2][8];

    uint32 **smsPads;
    uint32 **cvPads;
    uint32 **nesPads;
    NSMutableDictionary *smsPadPtrs;
    NSMutableDictionary *cvPadPtrs;
    NSMutableDictionary *nesPadPtrs;
}

+ (void)initialize;
+ (void)createAppSupportDirs;

- (IBAction)prefToggle:(id)sender;
- (IBAction)prefGGBiosBrowse:(id)sender;
- (IBAction)prefSMSPadButton:(id)sender;
- (IBAction)prefNESPadButton:(id)sender;
- (IBAction)prefCVPadButton:(id)sender;
- (IBAction)prefKeyboardReset:(id)sender;
- (IBAction)prefSaveDirBrowse:(id)sender;
- (IBAction)prefSaveStateDirBrowse:(id)sender;
- (IBAction)prefCheatDirBrowse:(id)sender;
- (IBAction)prefDirectoriesReset:(id)sender;
- (IBAction)prefSMSBiosBrowse:(id)sender;
- (IBAction)prefCVBiosBrowse:(id)sender;
- (IBAction)scanJoysticks:(id)sender;

- (void)awakeFromNib;
- (void)dealloc;

- (void)setSMSButton:(int)button
          forGamepad:(gamepad_t *)pad
              player:(int)player
          assignment:(uint32)a;
- (void)setCVButton:(int)button
         forGamepad:(gamepad_t *)pad
             player:(int)player
         assignment:(uint32)a;
- (void)setNESButton:(int)button
          forGamepad:(gamepad_t *)pad
              player:(int)player
          assignment:(uint32)a;
- (void)setButton:(int)button
       forGamepad:(gamepad_t *)pad
           player:(int)player
       assignment:(uint32)a
          console:(int)console;

- (unichar)keyValueForButton:(int)button
                      player:(int)player
                     console:(int)console;

- (gamepad_t *)getPadButton;
- (void)savePrefs;

- (BOOL)control:(NSControl *)c textShouldEndEditing:(NSText *)ed;

- (IBAction)buttonSelect:(id)sender;
- (IBAction)buttonSetOk:(id)sender;
- (IBAction)buttonSetCancel:(id)sender;

- (NSString *)keyListForConsole:(int)console
                         player:(int)player;

- (const uint32 *)gamepadAssignmentForConsole:(int)console
                                      gamepad:(gamepad_t *)pad;

- (BOOL)keyboardWithGamepadEnabled;

@end

#endif /* !CRABEMUPREFSCONTROLLER_H */
