/*
    This file is part of CrabEmu.

    Copyright (C) 2005, 2006, 2007, 2008, 2009, 2012 Lawrence Sebald

    CrabEmu is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2 
    as published by the Free Software Foundation.

    CrabEmu is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CrabEmu; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <unistd.h>

#include "CrabEmuPrefsController.h"
#include "EmuController.h"
#include "sound.h"

#include "sms.h"
#include "nes.h"
#include "colecovision.h"

@implementation CrabEmuPrefsController

+ (void)initialize
{
    NSUserDefaults *d = [NSUserDefaults standardUserDefaults];
    NSMutableDictionary *p = [NSMutableDictionary dictionary];
    NSMutableDictionary *km = [NSMutableDictionary dictionaryWithCapacity:2];
    NSMutableDictionary *kb;
    NSString *dir = @"~/Library/Application Support/ljsdcdev/CrabEmu";

    /* Make sure the Application Support directories exist. */
    [self createAppSupportDirs];

    /* Set up the defaults */
    [p setObject:[NSNumber numberWithBool:YES] forKey:@"PreserveAspectRatio"];
    [p setObject:[NSNumber numberWithBool:YES] forKey:@"DisableScreensaver"];
    [p setObject:[NSNumber numberWithBool:NO] forKey:@"StartFullscreen"];
    [p setObject:[NSNumber numberWithBool:NO] forKey:@"GameGearBiosEnabled"];
    [p setObject:[NSNumber numberWithBool:NO] forKey:@"SMSBiosEnabled"];
    [p setObject:[NSNumber numberWithBool:YES] forKey:@"SMSEnablePSG"];
    [p setObject:[NSNumber numberWithBool:YES] forKey:@"SMSEnableYM2413"];
    [p setObject:[NSNumber numberWithBool:YES] forKey:@"SMSPSGTone0"];
    [p setObject:[NSNumber numberWithBool:YES] forKey:@"SMSPSGTone1"];
    [p setObject:[NSNumber numberWithBool:YES] forKey:@"SMSPSGTone2"];
    [p setObject:[NSNumber numberWithBool:YES] forKey:@"SMSPSGNoise"];
    [p setObject:[NSNumber numberWithBool:YES] forKey:@"SMSDomestic"];
    [p setObject:[NSNumber numberWithBool:NO] forKey:@"SMSPAL"];
    [p setObject:[NSNumber numberWithBool:YES] forKey:@"EnableKeyboardGamepad"];

    /* Set up the default paths. */
    dir = [dir stringByStandardizingPath];
    [p setObject:[dir stringByAppendingPathComponent:@"SRAM Saves"]
          forKey:@"SaveDirectory"];
    [p setObject:[dir stringByAppendingPathComponent:@"Cheat Codes"]
          forKey:@"CheatDirectory"];
    [p setObject:[dir stringByAppendingPathComponent:@"Save States"]
          forKey:@"StateDirectory"];

    /* Set up the default keyboard controls. */
    kb = [NSMutableDictionary dictionaryWithCapacity:7];
    [kb setObject:[NSNumber numberWithInt:NSUpArrowFunctionKey]
           forKey:@"SMS Up"];
    [kb setObject:[NSNumber numberWithInt:NSDownArrowFunctionKey]
           forKey:@"SMS Down"];
    [kb setObject:[NSNumber numberWithInt:NSLeftArrowFunctionKey]
           forKey:@"SMS Left"];
    [kb setObject:[NSNumber numberWithInt:NSRightArrowFunctionKey]
           forKey:@"SMS Right"];
    [kb setObject:[NSNumber numberWithInt:';'] forKey:@"SMS Button 1"];
    [kb setObject:[NSNumber numberWithInt:'\''] forKey:@"SMS Button 2"];
    [kb setObject:[NSNumber numberWithInt:'p'] forKey:@"Game Gear Start"];
    [km setObject:kb forKey:@"Keyboard 1"];

    kb = [NSMutableDictionary dictionaryWithCapacity:7];
    [kb setObject:[NSNumber numberWithInt:'w'] forKey:@"SMS Up"];
    [kb setObject:[NSNumber numberWithInt:'s'] forKey:@"SMS Down"];
    [kb setObject:[NSNumber numberWithInt:'a'] forKey:@"SMS Left"];
    [kb setObject:[NSNumber numberWithInt:'d'] forKey:@"SMS Right"];
    [kb setObject:[NSNumber numberWithInt:'v'] forKey:@"SMS Button 1"];
    [kb setObject:[NSNumber numberWithInt:'b'] forKey:@"SMS Button 2"];
    [kb setObject:[NSNumber numberWithInt:'e'] forKey:@"Game Gear Start"];
    [km setObject:kb forKey:@"Keyboard 2"];
    [p setObject:km forKey:@"SMS Buttonmap"];

    km = [NSMutableDictionary dictionaryWithCapacity:2];
    kb = [NSMutableDictionary dictionaryWithCapacity:18];
    [kb setObject:[NSNumber numberWithInt:'w'] forKey:@"Up"];
    [kb setObject:[NSNumber numberWithInt:'a'] forKey:@"Left"];
    [kb setObject:[NSNumber numberWithInt:'s'] forKey:@"Down"];
    [kb setObject:[NSNumber numberWithInt:'d'] forKey:@"Right"];
    [kb setObject:[NSNumber numberWithInt:'q'] forKey:@"Left Action"];
    [kb setObject:[NSNumber numberWithInt:'e'] forKey:@"Right Action"];
    [kb setObject:[NSNumber numberWithInt:'4'] forKey:@"1"];
    [kb setObject:[NSNumber numberWithInt:'5'] forKey:@"2"];
    [kb setObject:[NSNumber numberWithInt:'6'] forKey:@"3"];
    [kb setObject:[NSNumber numberWithInt:'r'] forKey:@"4"];
    [kb setObject:[NSNumber numberWithInt:'t'] forKey:@"5"];
    [kb setObject:[NSNumber numberWithInt:'y'] forKey:@"6"];
    [kb setObject:[NSNumber numberWithInt:'f'] forKey:@"7"];
    [kb setObject:[NSNumber numberWithInt:'g'] forKey:@"8"];
    [kb setObject:[NSNumber numberWithInt:'h'] forKey:@"9"];
    [kb setObject:[NSNumber numberWithInt:'v'] forKey:@"*"];
    [kb setObject:[NSNumber numberWithInt:'b'] forKey:@"0"];
    [kb setObject:[NSNumber numberWithInt:'n'] forKey:@"#"];
    [km setObject:kb forKey:@"Keyboard 1"];

    kb = [NSMutableDictionary dictionaryWithCapacity:18];
    [kb setObject:[NSNumber numberWithInt:NSUpArrowFunctionKey] forKey:@"Up"];
    [kb setObject:[NSNumber numberWithInt:NSLeftArrowFunctionKey]
           forKey:@"Left"];
    [kb setObject:[NSNumber numberWithInt:NSDownArrowFunctionKey]
           forKey:@"Down"];
    [kb setObject:[NSNumber numberWithInt:NSRightArrowFunctionKey]
           forKey:@"Right"];
    [kb setObject:[NSNumber numberWithInt:'-'] forKey:@"Left Action"];
    [kb setObject:[NSNumber numberWithInt:'='] forKey:@"Right Action"];
    [kb setObject:[NSNumber numberWithInt:'8'] forKey:@"1"];
    [kb setObject:[NSNumber numberWithInt:'9'] forKey:@"2"];
    [kb setObject:[NSNumber numberWithInt:'0'] forKey:@"3"];
    [kb setObject:[NSNumber numberWithInt:'i'] forKey:@"4"];
    [kb setObject:[NSNumber numberWithInt:'o'] forKey:@"5"];
    [kb setObject:[NSNumber numberWithInt:'p'] forKey:@"6"];
    [kb setObject:[NSNumber numberWithInt:'k'] forKey:@"7"];
    [kb setObject:[NSNumber numberWithInt:'l'] forKey:@"8"];
    [kb setObject:[NSNumber numberWithInt:';'] forKey:@"9"];
    [kb setObject:[NSNumber numberWithInt:','] forKey:@"*"];
    [kb setObject:[NSNumber numberWithInt:'.'] forKey:@"0"];
    [kb setObject:[NSNumber numberWithInt:'/'] forKey:@"#"];
    [km setObject:kb forKey:@"Keyboard 2"];
    [p setObject:km forKey:@"ColecoVision Buttonmap"];

    kb = [NSMutableDictionary dictionaryWithCapacity:8];
    [kb setObject:[NSNumber numberWithInt:NSUpArrowFunctionKey]
           forKey:@"Up"];
    [kb setObject:[NSNumber numberWithInt:NSDownArrowFunctionKey]
           forKey:@"Down"];
    [kb setObject:[NSNumber numberWithInt:NSLeftArrowFunctionKey]
           forKey:@"Left"];
    [kb setObject:[NSNumber numberWithInt:NSRightArrowFunctionKey]
           forKey:@"Right"];
    [kb setObject:[NSNumber numberWithInt:'\''] forKey:@"A"];
    [kb setObject:[NSNumber numberWithInt:';'] forKey:@"B"];
    [kb setObject:[NSNumber numberWithInt:'p'] forKey:@"Start"];
    [kb setObject:[NSNumber numberWithInt:'o'] forKey:@"Select"];
    [km setObject:kb forKey:@"Keyboard 1"];
    
    kb = [NSMutableDictionary dictionaryWithCapacity:8];
    [kb setObject:[NSNumber numberWithInt:'w'] forKey:@"Up"];
    [kb setObject:[NSNumber numberWithInt:'s'] forKey:@"Down"];
    [kb setObject:[NSNumber numberWithInt:'a'] forKey:@"Left"];
    [kb setObject:[NSNumber numberWithInt:'d'] forKey:@"Right"];
    [kb setObject:[NSNumber numberWithInt:'b'] forKey:@"A"];
    [kb setObject:[NSNumber numberWithInt:'v'] forKey:@"B"];
    [kb setObject:[NSNumber numberWithInt:'e'] forKey:@"Start"];
    [kb setObject:[NSNumber numberWithInt:'q'] forKey:@"Select"];
    [km setObject:kb forKey:@"Keyboard 2"];
    [p setObject:km forKey:@"NES Buttonmap"];

    /* Set the preferences in the NSUserDefaults. These will only be used if the
       user hasn't already set their own preferences. */
    [d registerDefaults:p];
}

+ (void)createAppSupportDirs
{
    NSFileManager *fm = [NSFileManager defaultManager];
    NSString *dir, *dir2;

    /* Make sure ~/Library/Application Support/ljsdcdev exists first. */
    dir = @"~/Library/Application Support/ljsdcdev";
    dir = [dir stringByStandardizingPath];

    if([fm fileExistsAtPath:dir] == NO) {
        [fm createDirectoryAtPath:dir
                       attributes:nil];
    }

    /* Next, make sure the CrabEmu directory exists. */
    dir = [dir stringByAppendingPathComponent:@"CrabEmu"];

    if([fm fileExistsAtPath:dir] == NO) {
        [fm createDirectoryAtPath:dir
                       attributes:nil];
    }

    /* Next, the CrabEmu/SRAM Saves directory. */
    dir2 = [dir stringByAppendingPathComponent:@"SRAM Saves"];

    if([fm fileExistsAtPath:dir2] == NO) {
        [fm createDirectoryAtPath:dir2
                       attributes:nil];
    }

    /* Next, the CrabEmu/Cheat Codes directory. */
    dir2 = [dir stringByAppendingPathComponent:@"Cheat Codes"];

    if([fm fileExistsAtPath:dir2] == NO) {
        [fm createDirectoryAtPath:dir2
                       attributes:nil];
    }

    /* And finally, the CrabEmu/Save States directory. */
    dir2 = [dir stringByAppendingPathComponent:@"Save States"];

    if([fm fileExistsAtPath:dir2] == NO) {
        [fm createDirectoryAtPath:dir2
                       attributes:nil];
    }
}

- (IBAction)prefToggle:(id)sender
{
    if([sender tag] == 1) {
        [mainController toggleScreensaver:[sender state]];
    }
}

- (IBAction)prefGGBiosBrowse:(id)sender
{
    NSOpenPanel *openPanel = [NSOpenPanel openPanel];
    NSArray *fileTypes = [NSArray arrayWithObjects:@"gg", nil];
	[openPanel setTitle:@"Choose a Game Gear BIOS Rom"];
	[openPanel setCanChooseDirectories:NO];

	if([openPanel runModalForTypes:fileTypes] == NSOKButton) {
		NSString *fn = [openPanel filename];

        [ggBios setStringValue:fn];
        [prefs setObject:fn forKey:@"GameGearBiosFile"];
        [prefs synchronize];
    }
}

static uint32 button_data;

static void gp_cb(gamepad_t *pad, gp_elemdata_t *elem, int value) {
    int range;

    switch(elem->type) {
        case GP_TYPE_BUTTON:
            /* Ignore button release events... */
            if(value)
                button_data = elem->number;
            break;

        case GP_TYPE_HAT:
            /* Ignore hat centering events... */
            if(value)
                button_data = 0x80000000 | (value << 8) | elem->number;
            break;

        case GP_TYPE_X_AXIS:
        case GP_TYPE_Y_AXIS:
        case GP_TYPE_Z_AXIS:
        case GP_TYPE_RX_AXIS:
        case GP_TYPE_RY_AXIS:
        case GP_TYPE_RZ_AXIS:
            range = elem->max - elem->min + 1;
            value -= range / 2;

            if(value < -(range / 4))
                button_data = 0x60000000 | elem->type;
            else if(value > (range / 4))
                button_data = 0x40000000 | elem->type;
            break;
    }
}

- (gamepad_t *)getPadButton
{
    gamepad_t *gp = gp_get_gamepad([pjoysticks indexOfSelectedItem]);
    int was_open = gp->open;
    gp_callback old_cb = gp->cb;
    NSRunLoop *rl;

    /* Create the semaphore we'll need... */
    button_data = 0;

    /* Open the gamepad if needed. */
    if(!was_open)
        gp_open(gp, &gp_cb);
    else
        /* Re-assign the callback to us for now... */
        gp->cb = gp_cb;

    [NSApp beginSheet:joyPanel
       modalForWindow:prefWindow
        modalDelegate:self
       didEndSelector:NULL
          contextInfo:NULL];

    rl = [NSRunLoop currentRunLoop];

    while(!button_data && [rl runMode:NSDefaultRunLoopMode
                           beforeDate:[NSDate distantFuture]]) ;

    [NSApp endSheet:joyPanel];
    [joyPanel orderOut:nil];

    if(!was_open)
        gp_close(gp);
    else
        gp->cb = old_cb;

    return gp;
}

- (IBAction)prefSMSPadButton:(id)sender
{
    gamepad_t *gp = [self getPadButton];
    int tag = [sender tag];

    [self setSMSButton:tag forGamepad:gp player:1 assignment:button_data];
    [self savePrefs];
}

- (IBAction)prefNESPadButton:(id)sender
{
    gamepad_t *gp = [self getPadButton];
    int tag = [sender tag];

    [self setNESButton:tag forGamepad:gp player:1 assignment:button_data];
    [self savePrefs];
}

- (IBAction)prefCVPadButton:(id)sender
{
    gamepad_t *gp = [self getPadButton];
    int tag = [sender tag];

    [self setCVButton:tag forGamepad:gp player:1 assignment:button_data];
    [self savePrefs];
}

- (IBAction)prefKeyboardReset:(id)sender
{
    switch([sender tag]) {
        case 0:
            [self setSMSButton:SMS_UP forGamepad:NULL player:1
                    assignment:NSUpArrowFunctionKey];
            [self setSMSButton:SMS_DOWN forGamepad:NULL player:1
                    assignment:NSDownArrowFunctionKey];
            [self setSMSButton:SMS_LEFT forGamepad:NULL player:1
                    assignment:NSLeftArrowFunctionKey];
            [self setSMSButton:SMS_RIGHT forGamepad:NULL player:1
                    assignment:NSRightArrowFunctionKey];
            [self setSMSButton:SMS_BUTTON_1 forGamepad:NULL player:1
                    assignment:';'];
            [self setSMSButton:SMS_BUTTON_2 forGamepad:NULL player:1
                    assignment:'\''];
            [self setSMSButton:GAMEGEAR_START forGamepad:NULL player:1
                    assignment:'p'];
            break;

        case 1:
            [self setSMSButton:SMS_UP forGamepad:NULL player:2
                    assignment:'w'];
            [self setSMSButton:SMS_DOWN forGamepad:NULL player:2
                    assignment:'s'];
            [self setSMSButton:SMS_LEFT forGamepad:NULL player:2
                    assignment:'a'];
            [self setSMSButton:SMS_RIGHT forGamepad:NULL player:2
                    assignment:'d'];
            [self setSMSButton:SMS_BUTTON_1 forGamepad:NULL player:2
                    assignment:'v'];
            [self setSMSButton:SMS_BUTTON_2 forGamepad:NULL player:2
                    assignment:'b'];
            [self setSMSButton:GAMEGEAR_START forGamepad:NULL player:2
                    assignment:'e'];
            break;

        case 2:
            [self setCVButton:COLECOVISION_UP forGamepad:NULL player:1
                   assignment:'w'];
            [self setCVButton:COLECOVISION_DOWN forGamepad:NULL player:1
                   assignment:'s'];
            [self setCVButton:COLECOVISION_LEFT forGamepad:NULL player:1
                   assignment:'a'];
            [self setCVButton:COLECOVISION_RIGHT forGamepad:NULL player:1
                   assignment:'d'];
            [self setCVButton:COLECOVISION_L_ACTION forGamepad:NULL player:1
                   assignment:'q'];
            [self setCVButton:COLECOVISION_R_ACTION forGamepad:NULL player:1
                   assignment:'e'];
            [self setCVButton:COLECOVISION_1 forGamepad:NULL player:1
                   assignment:'4'];
            [self setCVButton:COLECOVISION_2 forGamepad:NULL player:1
                   assignment:'5'];
            [self setCVButton:COLECOVISION_3 forGamepad:NULL player:1
                   assignment:'6'];
            [self setCVButton:COLECOVISION_4 forGamepad:NULL player:1
                   assignment:'r'];
            [self setCVButton:COLECOVISION_5 forGamepad:NULL player:1
                   assignment:'t'];
            [self setCVButton:COLECOVISION_6 forGamepad:NULL player:1
                   assignment:'y'];
            [self setCVButton:COLECOVISION_7 forGamepad:NULL player:1
                   assignment:'f'];
            [self setCVButton:COLECOVISION_8 forGamepad:NULL player:1
                   assignment:'g'];
            [self setCVButton:COLECOVISION_9 forGamepad:NULL player:1
                   assignment:'h'];
            [self setCVButton:COLECOVISION_STAR forGamepad:NULL player:1
                   assignment:'v'];
            [self setCVButton:COLECOVISION_0 forGamepad:NULL player:1
                   assignment:'b'];
            [self setCVButton:COLECOVISION_POUND forGamepad:NULL player:1
                   assignment:'n'];
            break;

        case 3:
            [self setCVButton:COLECOVISION_UP forGamepad:NULL player:2
                   assignment:NSUpArrowFunctionKey];
            [self setCVButton:COLECOVISION_DOWN forGamepad:NULL player:2
                   assignment:NSDownArrowFunctionKey];
            [self setCVButton:COLECOVISION_LEFT forGamepad:NULL player:2
                   assignment:NSLeftArrowFunctionKey];
            [self setCVButton:COLECOVISION_RIGHT forGamepad:NULL player:2
                   assignment:NSRightArrowFunctionKey];
            [self setCVButton:COLECOVISION_L_ACTION forGamepad:NULL player:2
                   assignment:'-'];
            [self setCVButton:COLECOVISION_R_ACTION forGamepad:NULL player:2
                   assignment:'='];
            [self setCVButton:COLECOVISION_1 forGamepad:NULL player:2
                   assignment:'8'];
            [self setCVButton:COLECOVISION_2 forGamepad:NULL player:2
                   assignment:'9'];
            [self setCVButton:COLECOVISION_3 forGamepad:NULL player:2
                   assignment:'0'];
            [self setCVButton:COLECOVISION_4 forGamepad:NULL player:2
                   assignment:'i'];
            [self setCVButton:COLECOVISION_5 forGamepad:NULL player:2
                   assignment:'o'];
            [self setCVButton:COLECOVISION_6 forGamepad:NULL player:2
                   assignment:'p'];
            [self setCVButton:COLECOVISION_7 forGamepad:NULL player:2
                   assignment:'k'];
            [self setCVButton:COLECOVISION_8 forGamepad:NULL player:2
                   assignment:'l'];
            [self setCVButton:COLECOVISION_9 forGamepad:NULL player:2
                   assignment:';'];
            [self setCVButton:COLECOVISION_STAR forGamepad:NULL player:2
                   assignment:','];
            [self setCVButton:COLECOVISION_0 forGamepad:NULL player:2
                   assignment:'.'];
            [self setCVButton:COLECOVISION_POUND forGamepad:NULL player:2
                   assignment:'/'];
            break;

        case 4:
            [self setNESButton:NES_UP forGamepad:NULL player:1
                    assignment:NSUpArrowFunctionKey];
            [self setNESButton:NES_DOWN forGamepad:NULL player:1
                    assignment:NSDownArrowFunctionKey];
            [self setNESButton:NES_LEFT forGamepad:NULL player:1
                    assignment:NSLeftArrowFunctionKey];
            [self setNESButton:NES_RIGHT forGamepad:NULL player:1
                    assignment:NSRightArrowFunctionKey];
            [self setNESButton:NES_A forGamepad:NULL player:1
                    assignment:'\''];
            [self setNESButton:NES_B forGamepad:NULL player:1
                    assignment:';'];
            [self setNESButton:NES_START forGamepad:NULL player:1
                    assignment:'p'];
            [self setNESButton:NES_SELECT forGamepad:NULL player:1
                    assignment:'o'];
            break;
            
        case 5:
            [self setNESButton:NES_UP forGamepad:NULL player:2
                    assignment:'w'];
            [self setNESButton:NES_DOWN forGamepad:NULL player:2
                    assignment:'s'];
            [self setNESButton:NES_LEFT forGamepad:NULL player:2
                    assignment:'a'];
            [self setNESButton:NES_RIGHT forGamepad:NULL player:2
                    assignment:'d'];
            [self setNESButton:NES_A forGamepad:NULL player:2
                    assignment:'b'];
            [self setNESButton:NES_B forGamepad:NULL player:2
                    assignment:'v'];
            [self setNESButton:NES_START forGamepad:NULL player:2
                    assignment:'e'];
            [self setNESButton:NES_SELECT forGamepad:NULL player:2
                    assignment:'q'];
            break;
    }

    [self savePrefs];
}

- (IBAction)prefSaveDirBrowse:(id)sender
{
    NSOpenPanel *openPanel = [NSOpenPanel openPanel];
	[openPanel setTitle:@"Choose a Save Directory"];
	[openPanel setCanChooseDirectories:YES];
    [openPanel setCanChooseFiles:NO];

	if([openPanel runModalForTypes:nil] == NSOKButton) {
		NSString *fn = [openPanel filename];

        [saveDir setStringValue:fn];
        [prefs setObject:fn forKey:@"SaveDirectory"];
        [prefs synchronize];
    }
}

- (IBAction)prefSaveStateDirBrowse:(id)sender
{
    NSOpenPanel *openPanel = [NSOpenPanel openPanel];
	[openPanel setTitle:@"Choose a Save State Directory"];
	[openPanel setCanChooseDirectories:YES];
    [openPanel setCanChooseFiles:NO];

	if([openPanel runModalForTypes:nil] == NSOKButton) {
		NSString *fn = [openPanel filename];

        [saveStateDir setStringValue:fn];
        [prefs setObject:fn forKey:@"StateDirectory"];
        [prefs synchronize];
    }
}

- (IBAction)prefCheatDirBrowse:(id)sender
{
    NSOpenPanel *openPanel = [NSOpenPanel openPanel];
	[openPanel setTitle:@"Choose a Cheat Code Directory"];
	[openPanel setCanChooseDirectories:YES];
    [openPanel setCanChooseFiles:NO];

	if([openPanel runModalForTypes:nil] == NSOKButton) {
		NSString *fn = [openPanel filename];

        [cheatDir setStringValue:fn];
        [prefs setObject:fn forKey:@"CheatDirectory"];
        [prefs synchronize];
    }
}

- (IBAction)prefDirectoriesReset:(id)sender
{
    /* Remove the user-specified version of these. NSUserDefaults will then fall
       back on the default versions. */
    [prefs removeObjectForKey:@"SaveDirectory"];
    [prefs removeObjectForKey:@"StateDirectory"];
    [prefs removeObjectForKey:@"CheatDirectory"];
    [prefs synchronize];
}

- (IBAction)prefSMSBiosBrowse:(id)sender
{
    NSOpenPanel *openPanel = [NSOpenPanel openPanel];
    NSArray *fileTypes = [NSArray arrayWithObjects:@"sms", nil];
	[openPanel setTitle:@"Choose a Master System BIOS Rom"];
	[openPanel setCanChooseDirectories:NO];

	if([openPanel runModalForTypes:fileTypes] == NSOKButton) {
		NSString *fn = [openPanel filename];

        [smsBios setStringValue:fn];
        [prefs setObject:fn forKey:@"SMSBiosFile"];
        [prefs synchronize];
    }
}

- (IBAction)prefCVBiosBrowse:(id)sender
{
    NSOpenPanel *openPanel = [NSOpenPanel openPanel];
    NSArray *fileTypes = [NSArray arrayWithObjects:@"rom", nil];

	[openPanel setTitle:@"Choose a ColecoVision BIOS Rom"];
	[openPanel setCanChooseDirectories:NO];

	if([openPanel runModalForTypes:fileTypes] == NSOKButton) {
		NSString *fn = [openPanel filename];

        [cvBios setStringValue:fn];
        [prefs setObject:fn forKey:@"ColecoVisionBiosFile"];
        [prefs synchronize];
    }
}

- (IBAction)scanJoysticks:(id)sender
{
    gamepad_t *pad;
    NSString *str;
    NSMenuItem *item, *item2;
    int i;

    /* Release any already scanned in gamepads. */
    gp_cleanup();

    while([p1input numberOfItems] > 1) {
        [p1input removeItemAtIndex:1];
        [p2input removeItemAtIndex:1];
    }

    [pjoysticks removeAllItems];

    /* Scan for any gamepads attached to the system. */
    pad_count = gp_scan();

    if(!pad_count) {
        [pjoysticks addItemWithTitle:@"No Gamepads Detected"];
        [pjoysticks setEnabled:NO];
        [pjoyUp setEnabled:NO];
        [pjoyRight setEnabled:NO];
        [pjoyDown setEnabled:NO];
        [pjoyLeft setEnabled:NO];
        [pjoyB1 setEnabled:NO];
        [pjoyB2 setEnabled:NO];
        [pjoyGGStart setEnabled:NO];

        [pNESPadUp setEnabled:NO];
        [pNESPadDown setEnabled:NO];
        [pNESPadLeft setEnabled:NO];
        [pNESPadRight setEnabled:NO];
        [pNESPadStart setEnabled:NO];
        [pNESPadSelect setEnabled:NO];
        [pNESPadA setEnabled:NO];
        [pNESPadB setEnabled:NO];

        [pCVPadUp setEnabled:NO];
        [pCVPadDown setEnabled:NO];
        [pCVPadLeft setEnabled:NO];
        [pCVPadRight setEnabled:NO];
        [pCVPadLTrig setEnabled:NO];
        [pCVPadRTrig setEnabled:NO];
        [pCVPad1 setEnabled:NO];
        [pCVPad2 setEnabled:NO];
        [pCVPad3 setEnabled:NO];
        [pCVPad4 setEnabled:NO];
        [pCVPad5 setEnabled:NO];
        [pCVPad6 setEnabled:NO];
        [pCVPad7 setEnabled:NO];
        [pCVPad8 setEnabled:NO];
        [pCVPad9 setEnabled:NO];
        [pCVPadStar setEnabled:NO];
        [pCVPad0 setEnabled:NO];
        [pCVPadPound setEnabled:NO];
    }
    else {
        [pjoysticks setEnabled:YES];
        [pjoyUp setEnabled:YES];
        [pjoyRight setEnabled:YES];
        [pjoyDown setEnabled:YES];
        [pjoyLeft setEnabled:YES];
        [pjoyB1 setEnabled:YES];
        [pjoyB2 setEnabled:YES];
        [pjoyGGStart setEnabled:YES];

        [pNESPadUp setEnabled:YES];
        [pNESPadDown setEnabled:YES];
        [pNESPadLeft setEnabled:YES];
        [pNESPadRight setEnabled:YES];
        [pNESPadStart setEnabled:YES];
        [pNESPadSelect setEnabled:YES];
        [pNESPadA setEnabled:YES];
        [pNESPadB setEnabled:YES];

        [pCVPadUp setEnabled:YES];
        [pCVPadDown setEnabled:YES];
        [pCVPadLeft setEnabled:YES];
        [pCVPadRight setEnabled:YES];
        [pCVPadLTrig setEnabled:YES];
        [pCVPadRTrig setEnabled:YES];
        [pCVPad1 setEnabled:YES];
        [pCVPad2 setEnabled:YES];
        [pCVPad3 setEnabled:YES];
        [pCVPad4 setEnabled:YES];
        [pCVPad5 setEnabled:YES];
        [pCVPad6 setEnabled:YES];
        [pCVPad7 setEnabled:YES];
        [pCVPad8 setEnabled:YES];
        [pCVPad9 setEnabled:YES];
        [pCVPadStar setEnabled:YES];
        [pCVPad0 setEnabled:YES];
        [pCVPadPound setEnabled:YES];
    }

    for(i = 0; i < pad_count; ++i) {
        pad = gp_get_gamepad(i);
        str = [NSString stringWithUTF8String:pad->name];

        item = [[NSMenuItem alloc] initWithTitle:str
                                          action:@selector(setInput1:)
                                   keyEquivalent:@""];

        item2 = [[NSMenuItem alloc] initWithTitle:str
                                           action:@selector(setInput2:)
                                    keyEquivalent:@""];

        /* Set the tag on the menu items. */
        [item setTag:i];
        [item2 setTag:i];

        /* Set the target of the menu item to this object. */
        [item setTarget:mainController];
        [item2 setTarget:mainController];

        /* Add the items to the menus. */
        [p1input addItem:item];
        [p2input addItem:item2];

        /* Check if the device is configured. */
        if([keymap objectForKey:str] == nil) {
            [item setEnabled:NO];
            [item2 setEnabled:NO];
        }

        [item release];
        [item2 release];

        [pjoysticks addItemWithTitle:str];
    }
}

- (void)awakeFromNib
{
    NSDictionary *d;
    id obj;
    NSArray *arr;
    unsigned int i, j = 0;

    smsPadPtrs = [[NSMutableDictionary alloc] init];
    nesPadPtrs = [[NSMutableDictionary alloc] init];
    cvPadPtrs = [[NSMutableDictionary alloc] init];

    prefs = [[NSUserDefaults standardUserDefaults] retain];

    /* Grab the button mapping from the preferences. If there isn't one in the
       saved file, it'll grab the one we set up in the +initialize method. */
    keymap = [[NSMutableDictionary dictionaryWithDictionary:
               [prefs dictionaryForKey:@"SMS Buttonmap"]] retain];
    cvmap = [[NSMutableDictionary dictionaryWithDictionary:
              [prefs dictionaryForKey:@"ColecoVision Buttonmap"]] retain];
    nesmap = [[NSMutableDictionary dictionaryWithDictionary:
               [prefs dictionaryForKey:@"NES Buttonmap"]] retain];

    memset(smsButtons[0], 0xFF, sizeof(unichar) * 9);
    memset(smsButtons[1], 0xFF, sizeof(unichar) * 9);
    memset(cvButtons[0], 0xFF, sizeof(unichar) * 18);
    memset(cvButtons[1], 0xFF, sizeof(unichar) * 18);
    memset(nesButtons[0], 0xFF, sizeof(unichar) * 8);
    memset(nesButtons[0], 0xFF, sizeof(unichar) * 8);

    /* Build up the button list */
    if((d = [keymap objectForKey:@"Keyboard 1"])) {
        if((obj = [d objectForKey:@"SMS Up"]))
            smsButtons[0][SMS_UP] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"SMS Down"]))
            smsButtons[0][SMS_DOWN] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"SMS Left"]))
            smsButtons[0][SMS_LEFT] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"SMS Right"]))
            smsButtons[0][SMS_RIGHT] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"SMS Button 1"]))
            smsButtons[0][SMS_BUTTON_1] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"SMS Button 2"]))
            smsButtons[0][SMS_BUTTON_2] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"Game Gear Start"]))
            smsButtons[0][GAMEGEAR_START] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"Quit Emulator"]))
            smsButtons[0][SMS_QUIT] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"SMS Reset"]))
            smsButtons[0][SMS_CONSOLE_RESET] = (unichar)[obj unsignedLongValue];
    }

    if((d = [keymap objectForKey:@"Keyboard 2"])) {
        if((obj = [d objectForKey:@"SMS Up"]))
            smsButtons[1][SMS_UP] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"SMS Down"]))
            smsButtons[1][SMS_DOWN] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"SMS Left"]))
            smsButtons[1][SMS_LEFT] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"SMS Right"]))
            smsButtons[1][SMS_RIGHT] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"SMS Button 1"]))
            smsButtons[1][SMS_BUTTON_1] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"SMS Button 2"]))
            smsButtons[1][SMS_BUTTON_2] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"Game Gear Start"]))
            smsButtons[1][GAMEGEAR_START] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"Quit Emulator"]))
            smsButtons[1][SMS_QUIT] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"SMS Reset"]))
            smsButtons[1][SMS_CONSOLE_RESET] = (unichar)[obj unsignedLongValue];
    }

    if((d = [cvmap objectForKey:@"Keyboard 1"])) {
        if((obj = [d objectForKey:@"Up"]))
            cvButtons[0][COLECOVISION_UP] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"Down"]))
            cvButtons[0][COLECOVISION_DOWN] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"Left"]))
            cvButtons[0][COLECOVISION_LEFT] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"Right"]))
            cvButtons[0][COLECOVISION_RIGHT] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"Left Action"]))
            cvButtons[0][COLECOVISION_L_ACTION] =
                (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"Right Action"]))
            cvButtons[0][COLECOVISION_R_ACTION] =
                (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"1"]))
            cvButtons[0][COLECOVISION_1] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"2"]))
            cvButtons[0][COLECOVISION_2] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"3"]))
            cvButtons[0][COLECOVISION_3] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"4"]))
            cvButtons[0][COLECOVISION_4] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"5"]))
            cvButtons[0][COLECOVISION_5] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"6"]))
            cvButtons[0][COLECOVISION_6] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"7"]))
            cvButtons[0][COLECOVISION_7] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"8"]))
            cvButtons[0][COLECOVISION_8] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"9"]))
            cvButtons[0][COLECOVISION_9] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"*"]))
            cvButtons[0][COLECOVISION_STAR] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"0"]))
            cvButtons[0][COLECOVISION_0] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"#"]))
            cvButtons[0][COLECOVISION_POUND] = (unichar)[obj unsignedLongValue];
    }

    if((d = [cvmap objectForKey:@"Keyboard 2"])) {
        if((obj = [d objectForKey:@"Up"]))
            cvButtons[1][COLECOVISION_UP] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"Down"]))
            cvButtons[1][COLECOVISION_DOWN] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"Left"]))
            cvButtons[1][COLECOVISION_LEFT] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"Right"]))
            cvButtons[1][COLECOVISION_RIGHT] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"Left Action"]))
            cvButtons[1][COLECOVISION_L_ACTION] =
                (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"Right Action"]))
            cvButtons[1][COLECOVISION_R_ACTION] =
                (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"1"]))
            cvButtons[1][COLECOVISION_1] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"2"]))
            cvButtons[1][COLECOVISION_2] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"3"]))
            cvButtons[1][COLECOVISION_3] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"4"]))
            cvButtons[1][COLECOVISION_4] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"5"]))
            cvButtons[1][COLECOVISION_5] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"6"]))
            cvButtons[1][COLECOVISION_6] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"7"]))
            cvButtons[1][COLECOVISION_7] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"8"]))
            cvButtons[1][COLECOVISION_8] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"9"]))
            cvButtons[1][COLECOVISION_9] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"*"]))
            cvButtons[1][COLECOVISION_STAR] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"0"]))
            cvButtons[1][COLECOVISION_0] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"#"]))
            cvButtons[1][COLECOVISION_POUND] = (unichar)[obj unsignedLongValue];
    }

    if((d = [nesmap objectForKey:@"Keyboard 1"])) {
        if((obj = [d objectForKey:@"Up"]))
            nesButtons[0][NES_UP] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"Down"]))
            nesButtons[0][NES_DOWN] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"Left"]))
            nesButtons[0][NES_LEFT] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"Right"]))
            nesButtons[0][NES_RIGHT] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"A"]))
            nesButtons[0][NES_A] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"B"]))
            nesButtons[0][NES_B] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"Start"]))
            nesButtons[0][NES_START] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"Select"]))
            nesButtons[0][NES_SELECT] = (unichar)[obj unsignedLongValue];
    }

    if((d = [nesmap objectForKey:@"Keyboard 2"])) {
        if((obj = [d objectForKey:@"Up"]))
            nesButtons[1][NES_UP] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"Down"]))
            nesButtons[1][NES_DOWN] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"Left"]))
            nesButtons[1][NES_LEFT] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"Right"]))
            nesButtons[1][NES_RIGHT] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"A"]))
            nesButtons[1][NES_A] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"B"]))
            nesButtons[1][NES_B] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"Start"]))
            nesButtons[1][NES_START] = (unichar)[obj unsignedLongValue];
        if((obj = [d objectForKey:@"Select"]))
            nesButtons[1][NES_SELECT] = (unichar)[obj unsignedLongValue];
    }

    /* Set up the pad data arrays... */
    arr = [keymap allKeys];

    if([arr count] > 2) {
        smsPads = (uint32 **)malloc(([arr count] - 2) * sizeof(uint32 *));

        for(i = 0; i < [arr count]; ++i) {
            obj = [arr objectAtIndex:i];
            if([obj isEqualToString:@"Keyboard 1"] ||
               [obj isEqualToString:@"Keyboard 2"])
                continue;

            [smsPadPtrs setObject:[NSNumber numberWithInt:j] forKey:obj];
            smsPads[j] = (uint32 *)malloc(8 * sizeof(uint32));

            d = [keymap objectForKey:obj];
            if((obj = [d objectForKey:@"SMS Up"]))
                smsPads[j][SMS_UP] = (uint32)[obj unsignedLongValue];
            if((obj = [d objectForKey:@"SMS Down"]))
                smsPads[j][SMS_DOWN] = (uint32)[obj unsignedLongValue];
            if((obj = [d objectForKey:@"SMS Left"]))
                smsPads[j][SMS_LEFT] = (uint32)[obj unsignedLongValue];
            if((obj = [d objectForKey:@"SMS Right"]))
                smsPads[j][SMS_RIGHT] = (uint32)[obj unsignedLongValue];
            if((obj = [d objectForKey:@"SMS Button 1"]))
                smsPads[j][SMS_BUTTON_1] = (uint32)[obj unsignedLongValue];
            if((obj = [d objectForKey:@"SMS Button 2"]))
                smsPads[j][SMS_BUTTON_2] = (uint32)[obj unsignedLongValue];
            if((obj = [d objectForKey:@"Game Gear Start"]))
                smsPads[j][GAMEGEAR_START] = (uint32)[obj unsignedLongValue];
        }
    }

    arr = [nesmap allKeys];
    
    if([arr count] > 2) {
        nesPads = (uint32 **)malloc(([arr count] - 2) * sizeof(uint32 *));

        for(i = 0; i < [arr count]; ++i) {
            obj = [arr objectAtIndex:i];
            if([obj isEqualToString:@"Keyboard 1"] ||
               [obj isEqualToString:@"Keyboard 2"])
                continue;

            [nesPadPtrs setObject:[NSNumber numberWithInt:j] forKey:obj];
            nesPads[j] = (uint32 *)malloc(8 * sizeof(uint32));
            
            d = [nesmap objectForKey:obj];
            if((obj = [d objectForKey:@"Up"]))
                nesPads[j][NES_UP] = (uint32)[obj unsignedLongValue];
            if((obj = [d objectForKey:@"Down"]))
                nesPads[j][NES_DOWN] = (uint32)[obj unsignedLongValue];
            if((obj = [d objectForKey:@"Left"]))
                nesPads[j][NES_LEFT] = (uint32)[obj unsignedLongValue];
            if((obj = [d objectForKey:@"Right"]))
                nesPads[j][NES_RIGHT] = (uint32)[obj unsignedLongValue];
            if((obj = [d objectForKey:@"A"]))
                nesPads[j][NES_A] = (uint32)[obj unsignedLongValue];
            if((obj = [d objectForKey:@"B"]))
                nesPads[j][NES_B] = (uint32)[obj unsignedLongValue];
            if((obj = [d objectForKey:@"Start"]))
                nesPads[j][NES_START] = (uint32)[obj unsignedLongValue];
            if((obj = [d objectForKey:@"Select"]))
                nesPads[j][NES_SELECT] = (uint32)[obj unsignedLongValue];
        }
    }

    arr = [cvmap allKeys];

    if([arr count] > 2) {
        cvPads = (uint32 **)malloc(([arr count] - 2) * sizeof(uint32 *));

        for(i = 0; i < [arr count]; ++i) {
            obj = [arr objectAtIndex:i];
            if([obj isEqualToString:@"Keyboard 1"] ||
               [obj isEqualToString:@"Keyboard 2"])
                continue;

            [cvPadPtrs setObject:[NSNumber numberWithInt:j] forKey:obj];
            cvPads[j] = (uint32 *)malloc(8 * sizeof(uint32));

            d = [cvmap objectForKey:obj];
            if((obj = [d objectForKey:@"Up"]))
                cvPads[j][COLECOVISION_UP] = (uint32)[obj unsignedLongValue];
            if((obj = [d objectForKey:@"Down"]))
                cvPads[j][COLECOVISION_DOWN] = (uint32)[obj unsignedLongValue];
            if((obj = [d objectForKey:@"Left"]))
                cvPads[j][COLECOVISION_LEFT] = (uint32)[obj unsignedLongValue];
            if((obj = [d objectForKey:@"Right"]))
                cvPads[j][COLECOVISION_RIGHT] = (uint32)[obj unsignedLongValue];
            if((obj = [d objectForKey:@"Left Action"]))
                cvPads[j][COLECOVISION_L_ACTION] =
                    (uint32)[obj unsignedLongValue];
            if((obj = [d objectForKey:@"Right Action"]))
                cvPads[j][COLECOVISION_R_ACTION] =
                    (uint32)[obj unsignedLongValue];
            if((obj = [d objectForKey:@"1"]))
                cvPads[j][COLECOVISION_1] = (uint32)[obj unsignedLongValue];
            if((obj = [d objectForKey:@"2"]))
                cvPads[j][COLECOVISION_2] = (uint32)[obj unsignedLongValue];
            if((obj = [d objectForKey:@"3"]))
                cvPads[j][COLECOVISION_3] = (uint32)[obj unsignedLongValue];
            if((obj = [d objectForKey:@"4"]))
                cvPads[j][COLECOVISION_4] = (uint32)[obj unsignedLongValue];
            if((obj = [d objectForKey:@"5"]))
                cvPads[j][COLECOVISION_5] = (uint32)[obj unsignedLongValue];
            if((obj = [d objectForKey:@"6"]))
                cvPads[j][COLECOVISION_6] = (uint32)[obj unsignedLongValue];
            if((obj = [d objectForKey:@"7"]))
                cvPads[j][COLECOVISION_7] = (uint32)[obj unsignedLongValue];
            if((obj = [d objectForKey:@"8"]))
                cvPads[j][COLECOVISION_8] = (uint32)[obj unsignedLongValue];
            if((obj = [d objectForKey:@"9"]))
                cvPads[j][COLECOVISION_9] = (uint32)[obj unsignedLongValue];
            if((obj = [d objectForKey:@"*"]))
                cvPads[j][COLECOVISION_STAR] = (uint32)[obj unsignedLongValue];
            if((obj = [d objectForKey:@"0"]))
                cvPads[j][COLECOVISION_0] = (uint32)[obj unsignedLongValue];
            if((obj = [d objectForKey:@"#"]))
                cvPads[j][COLECOVISION_POUND] = (uint32)[obj unsignedLongValue];
        }
    }
}

- (void)dealloc
{
    int i, cnt;

    cnt = (int)[smsPadPtrs count];
    for(i = 0; i < cnt; ++i) {
        free(smsPads[i]);
    }

    cnt = (int)[nesPadPtrs count];
    for(i = 0; i < cnt; ++i) {
        free(nesPads[i]);
    }

    cnt = (int)[cvPadPtrs count];
    for(i = 0; i < cnt; ++i) {
        free(cvPads[i]);
    }

    free(smsPads);
    free(nesPads);
    free(cvPads);

    [smsPadPtrs release];
    [nesPadPtrs release];
    [cvPadPtrs release];
    
    [prefs release];
    [keymap release];
    [super dealloc];
}

- (void)setSMSButton:(int)button
         forGamepad:(gamepad_t *)pad
              player:(int)player
          assignment:(uint32)a
{
    NSString *str;
    NSMutableDictionary *d;
    NSNumber *n;
    int cnt;
    void *tmp;

    if(pad) {
        str = [NSString stringWithUTF8String:pad->name];
        n = [smsPadPtrs objectForKey:str];

        if(n) {
            smsPads[[n intValue]][button] = a;
        }
        else {
            /* Add a new set to our collection... */
            cnt = [smsPadPtrs count];
            if(!(tmp = realloc(smsPads, (cnt + 1) * sizeof(uint32 *))))
                /* XXXX */
                return;

            smsPads = (uint32 **)tmp;
            if(!(smsPads[cnt] = (uint32 *)malloc(sizeof(uint32) * 8)))
                /* XXXX */
                return;

            memset(smsPads[cnt], 0xFF, sizeof(uint32) * 8);
            smsPads[cnt][button] = a;
            [smsPadPtrs setObject:[NSNumber numberWithInt:cnt]
                           forKey:str];
            [self scanJoysticks:nil];
        }
    }
    else {
        str = [NSString stringWithFormat:@"Keyboard %d", player];
        smsButtons[player - 1][button] = (unichar)a;
    }

    if((d = [NSMutableDictionary dictionaryWithDictionary:
             [keymap objectForKey:str]]) == nil) {
        d = [NSMutableDictionary dictionaryWithCapacity:8];
    }

    switch(button) {
        case SMS_UP:
            [d setObject:[NSNumber numberWithUnsignedLong:a]
                  forKey:@"SMS Up"];
            break;

        case SMS_DOWN:
            [d setObject:[NSNumber numberWithUnsignedLong:a]
                  forKey:@"SMS Down"];
            break;

        case SMS_LEFT:
            [d setObject:[NSNumber numberWithUnsignedLong:a]
                  forKey:@"SMS Left"];
            break;

        case SMS_RIGHT:
            [d setObject:[NSNumber numberWithUnsignedLong:a]
                  forKey:@"SMS Right"];
            break;

        case SMS_BUTTON_1:
            [d setObject:[NSNumber numberWithUnsignedLong:a]
                  forKey:@"SMS Button 1"];
            break;

        case SMS_BUTTON_2:
            [d setObject:[NSNumber numberWithUnsignedLong:a]
                  forKey:@"SMS Button 2"];
            break;

        case GAMEGEAR_START:
            [d setObject:[NSNumber numberWithUnsignedLong:a]
                  forKey:@"Game Gear Start"];
            break;

        case SMS_QUIT:
            [d setObject:[NSNumber numberWithUnsignedLong:a]
                  forKey:@"Quit Emulator"];
            break;

        case SMS_CONSOLE_RESET:
            [d setObject:[NSNumber numberWithUnsignedLong:a]
                  forKey:@"SMS Reset"];
            break;
    }

    /* Reset the object on the keymap... */
    [keymap setObject:d forKey:str];
}

- (void)setCVButton:(int)button
        forGamepad:(gamepad_t *)pad
             player:(int)player
         assignment:(uint32)a
{
    NSString *str;
    NSMutableDictionary *d;
    NSNumber *n;
    int cnt;
    void *tmp;

    if(pad) {
        str = [NSString stringWithUTF8String:pad->name];
        n = [cvPadPtrs objectForKey:str];

        if(n) {
            cvPads[[n intValue]][button] = a;
        }
        else {
            /* Add a new set to our collection... */
            cnt = [cvPadPtrs count];
            if(!(tmp = realloc(cvPads, (cnt + 1) * sizeof(uint32 *))))
            /* XXXX */
                return;

            cvPads = (uint32 **)tmp;
            if(!(cvPads[cnt] = (uint32 *)malloc(sizeof(uint32) * 8)))
            /* XXXX */
                return;

            memset(cvPads[cnt], 0xFF, sizeof(uint32) * 8);
            cvPads[cnt][button] = a;
            [cvPadPtrs setObject:[NSNumber numberWithInt:cnt]
                           forKey:str];
            [self scanJoysticks:nil];
        }
    }
    else {
        str = [NSString stringWithFormat:@"Keyboard %d", player];
        cvButtons[player - 1][button] = (unichar)a;
    }

    if((d = [NSMutableDictionary dictionaryWithDictionary:
             [cvmap objectForKey:str]]) == nil) {
        d = [NSMutableDictionary dictionaryWithCapacity:8];
    }

    switch(button) {
        case COLECOVISION_UP:
            [d setObject:[NSNumber numberWithUnsignedLong:a]
                  forKey:@"Up"];
            break;

        case COLECOVISION_DOWN:
            [d setObject:[NSNumber numberWithUnsignedLong:a]
                  forKey:@"Down"];
            break;

        case COLECOVISION_LEFT:
            [d setObject:[NSNumber numberWithUnsignedLong:a]
                  forKey:@"Left"];
            break;

        case COLECOVISION_RIGHT:
            [d setObject:[NSNumber numberWithUnsignedLong:a]
                  forKey:@"Right"];
            break;

        case COLECOVISION_L_ACTION:
            [d setObject:[NSNumber numberWithUnsignedLong:a]
                  forKey:@"Left Action"];
            break;

        case COLECOVISION_R_ACTION:
            [d setObject:[NSNumber numberWithUnsignedLong:a]
                  forKey:@"Right Action"];
            break;

        case COLECOVISION_1:
            [d setObject:[NSNumber numberWithUnsignedLong:a]
                  forKey:@"1"];
            break;

        case COLECOVISION_2:
            [d setObject:[NSNumber numberWithUnsignedLong:a]
                  forKey:@"2"];
            break;

        case COLECOVISION_3:
            [d setObject:[NSNumber numberWithUnsignedLong:a]
                  forKey:@"3"];
            break;

        case COLECOVISION_4:
            [d setObject:[NSNumber numberWithUnsignedLong:a]
                  forKey:@"4"];
            break;

        case COLECOVISION_5:
            [d setObject:[NSNumber numberWithUnsignedLong:a]
                  forKey:@"5"];
            break;

        case COLECOVISION_6:
            [d setObject:[NSNumber numberWithUnsignedLong:a]
                  forKey:@"6"];
            break;

        case COLECOVISION_7:
            [d setObject:[NSNumber numberWithUnsignedLong:a]
                  forKey:@"7"];
            break;

        case COLECOVISION_8:
            [d setObject:[NSNumber numberWithUnsignedLong:a]
                  forKey:@"8"];
            break;

        case COLECOVISION_9:
            [d setObject:[NSNumber numberWithUnsignedLong:a]
                  forKey:@"9"];
            break;

        case COLECOVISION_STAR:
            [d setObject:[NSNumber numberWithUnsignedLong:a]
                  forKey:@"*"];
            break;

        case COLECOVISION_0:
            [d setObject:[NSNumber numberWithUnsignedLong:a]
                  forKey:@"0"];
            break;

        case COLECOVISION_POUND:
            [d setObject:[NSNumber numberWithUnsignedLong:a]
                  forKey:@"#"];
            break;
    }

    /* Reset the object on the keymap... */
    [cvmap setObject:d forKey:str];
}

- (void)setNESButton:(int)button
         forGamepad:(gamepad_t *)pad
              player:(int)player
          assignment:(uint32)a
{
    NSString *str;
    NSMutableDictionary *d;
    NSNumber *n;
    int cnt;
    void *tmp;

    if(pad) {
        str = [NSString stringWithUTF8String:pad->name];
        n = [nesPadPtrs objectForKey:str];

        if(n) {
            nesPads[[n intValue]][button] = a;
        }
        else {
            /* Add a new set to our collection... */
            cnt = [nesPadPtrs count];
            if(!(tmp = realloc(nesPads, (cnt + 1) * sizeof(uint32 *))))
            /* XXXX */
                return;

            nesPads = (uint32 **)tmp;
            if(!(nesPads[cnt] = (uint32 *)malloc(sizeof(uint32) * 8)))
            /* XXXX */
                return;

            memset(nesPads[cnt], 0xFF, sizeof(uint32) * 8);
            nesPads[cnt][button] = a;
            [nesPadPtrs setObject:[NSNumber numberWithInt:cnt]
                           forKey:str];
            [self scanJoysticks:nil];
        }
    }
    else {
        str = [NSString stringWithFormat:@"Keyboard %d", player];
        nesButtons[player - 1][button] = (unichar)a;
    }

    if((d = [NSMutableDictionary dictionaryWithDictionary:
             [keymap objectForKey:str]]) == nil) {
        d = [NSMutableDictionary dictionaryWithCapacity:8];
    }

    switch(button) {
        case NES_UP:
            [d setObject:[NSNumber numberWithUnsignedLong:a]
                  forKey:@"Up"];
            break;

        case NES_DOWN:
            [d setObject:[NSNumber numberWithUnsignedLong:a]
                  forKey:@"Down"];
            break;

        case NES_LEFT:
            [d setObject:[NSNumber numberWithUnsignedLong:a]
                  forKey:@"Left"];
            break;

        case NES_RIGHT:
            [d setObject:[NSNumber numberWithUnsignedLong:a]
                  forKey:@"Right"];
            break;

        case NES_A:
            [d setObject:[NSNumber numberWithUnsignedLong:a]
                  forKey:@"A"];
            break;

        case NES_B:
            [d setObject:[NSNumber numberWithUnsignedLong:a]
                  forKey:@"B"];
            break;

        case NES_START:
            [d setObject:[NSNumber numberWithUnsignedLong:a]
                  forKey:@"Start"];
            break;

        case NES_SELECT:
            [d setObject:[NSNumber numberWithUnsignedLong:a]
                  forKey:@"Select"];
            break;
    }

    /* Reset the object on the keymap... */
    [keymap setObject:d forKey:str];
}

- (void)setButton:(int)button
      forGamepad:(gamepad_t *)pad
           player:(int)player
       assignment:(uint32)a
          console:(int)console
{
    switch(console) {
        case CONSOLE_SMS:
        case CONSOLE_SG1000:
        case CONSOLE_SC3000:
        case CONSOLE_GG:
            [self setSMSButton:button
                   forGamepad:pad
                        player:player
                    assignment:a];
            break;

        case CONSOLE_COLECOVISION:
            [self setCVButton:button
                  forGamepad:pad
                       player:player
                   assignment:a];
            break;

        case CONSOLE_NES:
            [self setNESButton:button
                   forGamepad:pad
                        player:player
                    assignment:a];
            break;
    }
}

- (void)savePrefs
{
    /* Save the button mapping for the keyboard. */
    [prefs setObject:keymap forKey:@"SMS Buttonmap"];
    [prefs setObject:cvmap forKey:@"ColecoVision Buttonmap"];
    [prefs setObject:nesmap forKey:@"NES Buttonmap"];

    [prefs synchronize];
    [mainController readPrefs];
}

- (NSString *)prefsValueForControl:(NSControl *)c
{
    if(c == saveDir) {
        return [prefs stringForKey:@"SaveDirectory"];
    }
    else if(c == saveStateDir) {
        return [prefs stringForKey:@"StateDirectory"];
    }
    else if(c == cheatDir) {
        return [prefs stringForKey:@"CheatDirectory"];
    }
    else if(c == smsBios) {
        return [prefs stringForKey:@"SMSBiosFile"];
    }
    else if(c == ggBios) {
        return [prefs stringForKey:@"GameGearBiosFile"];
    }
    else if(c == cvBios) {
        return [prefs stringForKey:@"ColecoVisionBiosFile"];
    }

    return nil;
}

- (BOOL)control:(NSControl *)c textShouldEndEditing:(NSText *)ed
{
    NSString *s = [[ed string] stringByStandardizingPath];
    NSFileManager *fm = [NSFileManager defaultManager];
    BOOL dir = NO;
    int val = NSAlertDefaultReturn;
    BOOL err = NO;

    /* Update the value with the standardized version. */
    [ed setString:s];

    /* For the BIOS files, accept a blank path. This has the effect of disabling
       the selected BIOS. */
    if((c == smsBios || c == ggBios || c == cvBios) && [s length] == 0) {
        return YES;
    }

    /* Make sure the user gave us an absolute path. */
    if(![s isAbsolutePath]) {
        val = NSRunAlertPanel(@"Error", @"You must supply an absolute path.",
                              @"OK", @"Restore Previous Value", nil);
        err = YES;
    }

    /* Make sure the user gave us a path that exists. */
    if(![fm fileExistsAtPath:s isDirectory:&dir]) {
        val = NSRunAlertPanel(@"Error", @"Path at specified location does not "
                              @"exist.", @"OK", @"Restore Previous Value", nil);
        err = YES;
    }

    /* Make sure we can read the path. */
    if(![fm isReadableFileAtPath:s]) {
        val = NSRunAlertPanel(@"Error", @"Path at specified location is not "
                              @"readable.", @"OK", @"Restore Previous Value",
                              nil);
        err = YES;
    }

    /* Figure out what field the user is editing */
    if(c == saveDir || c == saveStateDir || c == cheatDir) {
        /* Make sure it is a directory. */
        if(!dir) {
            val = NSRunAlertPanel(@"Error", @"Path at specified location is "
                                  @"not a directory.", @"OK",
                                  @"Restore Previous Value", nil);
            err = YES;
        }

        /* Make sure we can write at the path. */
        if(![fm isWritableFileAtPath:s]) {
            val = NSRunAlertPanel(@"Error", @"Path at specified location is "
                                  @"not writable.", @"OK",
                                  @"Restore Previous Value", nil);
            err = YES;
        }
    }
    else {
        /* Make sure its not a directory. */
        if(dir) {
            val = NSRunAlertPanel(@"Error", @"Path at specified location is a "
                                  @"directory.", @"OK",
                                  @"Restore Previous Value", nil);
            err = YES;
        }
    }

    /* Were there any errors? */
    if(err) {
        if(val == NSAlertAlternateReturn) {
            /* Restore old value... */
            [ed setString:[self prefsValueForControl:c]];
            return YES;
        }
        else {
            return NO;
        }
    }

    return YES;
}

- (unichar)keyValueForButton:(int)button
                      player:(int)player
                     console:(int)console
{
    if(player < 1 || player > 2)
        return (unichar)-1;

    switch(console) {
        case CONSOLE_SMS:
        case CONSOLE_SG1000:
        case CONSOLE_SC3000:
        case CONSOLE_GG:
            if(button < 0 || button > 7)
                return (unichar)-1;

            return smsButtons[player - 1][button];

        case CONSOLE_COLECOVISION:
            if(button < 0 || button > 17)
                return (unichar)-1;

            return cvButtons[player - 1][button];

        case CONSOLE_NES:
            if(button < 0 || button > 8)
                return (unichar)-1;

            return nesButtons[player - 1][button];
    }

    return (unichar)-1;
}

- (IBAction)buttonSelect:(id)sender
{
    int rv;
    int tag = [sender tag];
    int console = (tag >> 16) & 0xFF;
    int player = (tag >> 8) & 0xFF;
    int num = tag & 0xFF;
    uint32 value = [self keyValueForButton:num player:player console:console];
    unichar ch;

    /* Fill in current setting from prefs */
    if(value != (uint32)-1) {
        switch(value) {
            case '\r':
                ch = 0x23CE;
                break;

            case '\t':
                ch = 0x21E5;
                break;

            case 27:
                ch = 0x241B;
                break;

            case 127:
                ch = 0x232B;
                break;

            case NSLeftArrowFunctionKey:
                ch = 0x2190;
                break;

            case NSUpArrowFunctionKey:
                ch = 0x2191;
                break;

            case NSRightArrowFunctionKey:
                ch = 0x2192;
                break;

            case NSDownArrowFunctionKey:
                ch = 0x2193;
                break;

            default:
                ch = toupper(((int)value));
        }

        [buttonBox setStringValue:[NSString stringWithCharacters:&ch length:1]];
    }
    else {
        [buttonBox setStringValue:@""];
    }

    /* Open up the sheet and ask for the user's input */
    [NSApp beginSheet:buttonAssignment
       modalForWindow:prefWindow
        modalDelegate:self
       didEndSelector:nil
          contextInfo:NULL];

    rv = [NSApp runModalForWindow:buttonAssignment];
    [NSApp endSheet:buttonAssignment];
    [buttonAssignment orderOut:nil];

    /* Did the user accept what they put in? */
    if(rv == NSOKButton) {
        NSString *s = [buttonBox stringValue];
        uint32 val;

        /* This shouldn't happen... */
        if([s length] < 1) {
            return;
        }

        switch([s characterAtIndex:0]) {
            case 0x23CE:    /* Return */
                val = '\r';
                break;

            case 0x21E5:    /* Tab */
                val = '\t';
                break;

            case 0x241B:    /* Escape */
                val = 27;
                break;

            case 0x232B:    /* Backspace */
                val = 127;
                break;

            case 0x2190:    /* Left */
                val = NSLeftArrowFunctionKey;
                break;

            case 0x2191:    /* Up */
                val = NSUpArrowFunctionKey;
                break;

            case 0x2192:    /* Right */
                val = NSRightArrowFunctionKey;
                break;

            case 0x2193:    /* Down */
                val = NSDownArrowFunctionKey;
                break;

            default:
                val = tolower([s characterAtIndex:0]);
        }

        /* Update the key mapping. This will also save the key to the
           preferences. */
        [self setButton:num
            forGamepad:NULL
                 player:player
             assignment:(uint32)val
                console:console];
        [self savePrefs];
    }
}

- (IBAction)buttonSetOk:(id)sender
{
    [NSApp stopModalWithCode:NSOKButton];
}

- (IBAction)buttonSetCancel:(id)sender
{
    [NSApp stopModalWithCode:NSCancelButton];
}

- (NSString *)keyListForConsole:(int)console
                         player:(int)player
{
    /* We don't support more than two controllers at the moment. */
    if(player < 1 || player > 2)
        return [NSString string];

    switch(console) {
        case CONSOLE_SMS:
        case CONSOLE_GG:
        case CONSOLE_SG1000:
        case CONSOLE_SC3000:
            return [NSString stringWithCharacters:smsButtons[player - 1]
                                           length:9];

        case CONSOLE_COLECOVISION:
            return [NSString stringWithCharacters:cvButtons[player - 1]
                                           length:18];

        case CONSOLE_NES:
            return [NSString stringWithCharacters:nesButtons[player - 1]
                                           length:8];
    }

    /* Don't have it... Punt. */
    return [NSString string];
}

- (const uint32 *)gamepadAssignmentForConsole:(int)console
                                      gamepad:(gamepad_t *)pad
{
    NSString *str = [NSString stringWithUTF8String:pad->name];
    NSNumber *idx;

    switch(console) {
        case CONSOLE_SMS:
        case CONSOLE_GG:
        case CONSOLE_SG1000:
        case CONSOLE_SC3000:
            if(!(idx = [smsPadPtrs objectForKey:str]))
                return NULL;

            return smsPads[[idx intValue]];

        case CONSOLE_COLECOVISION:
            if(!(idx = [cvPadPtrs objectForKey:str]))
                return NULL;

            return cvPads[[idx intValue]];

        case CONSOLE_NES:
            if(!(idx = [nesPadPtrs objectForKey:str]))
                return NULL;

            return nesPads[[idx intValue]];
    }

    /* Don't have it... Punt. */
    return NULL;
}

- (BOOL)keyboardWithGamepadEnabled
{
    return [keyboardWithGamepad state] == NSOnState;
}

@end
