/*
	This file is part of CrabEmu.

	Copyright (C) 2005, 2006, 2007, 2008, 2009, 2012 Lawrence Sebald

    CrabEmu is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2 
    as published by the Free Software Foundation.

	CrabEmu is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with CrabEmu; if not, write to the Free Software
	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#ifndef EMUCONTROLLER_H
#define EMUCONTROLLER_H

#import <Cocoa/Cocoa.h>
#include "CrabEmu.h"
#include "PaletteGLView.h"
#include "PatternGLView.h"
#include "NESPatternGLView.h"
#include "SDSCView.h"
#include "SMSCheatController.h"
#include "CrabEmuPrefsController.h"
#include "gamepad.h"

@interface EmuController : NSObject
{
    IBOutlet NSMenuItem *noise;
    IBOutlet NSMenuItem *pause;
    IBOutlet NSMenuItem *reset;
    IBOutlet NSMenuItem *PSG;
    IBOutlet NSMenuItem *YM2413;
    IBOutlet NSMenuItem *tone0;
    IBOutlet NSMenuItem *tone1;
    IBOutlet NSMenuItem *tone2;
    IBOutlet NSMenuItem *debugger;
    IBOutlet NSWindow *debuggerWindow;
    IBOutlet NSWindow *smsVdpDebuggerWindow;
    IBOutlet NSMenuItem *rExport;
    IBOutlet NSMenuItem *rDomestic;
    IBOutlet NSMenuItem *pal;
    IBOutlet NSMenuItem *ntsc;
    IBOutlet NSMenu *p1input;
    IBOutlet NSMenu *p2input;
    IBOutlet NSMenuItem *p1keyboard;
    IBOutlet NSMenuItem *p2keyboard;

    IBOutlet NSTextField *smsBios;
    IBOutlet NSButton *smsBiosEnabled;
    IBOutlet NSTextField *ggBios;
    IBOutlet NSButton *ggBiosEnabled;
    IBOutlet NSTextField *cvBios;
    IBOutlet NSTextField *saveDir;
    IBOutlet NSTextField *saveStateDir;
    IBOutlet NSTextField *cheatDir;
    IBOutlet NSButton *ssaverDisabled;
    IBOutlet NSButton *startFullscreen;
    IBOutlet NSButton *preserveAspect;

    IBOutlet NSMenu *saveStates;
    IBOutlet NSMenu *loadStates;
    IBOutlet NSMenuItem *saveSRAM;
    IBOutlet NSMenuItem *screenshot;
    IBOutlet NSMenuItem *fullscreen;
    IBOutlet NSMenuItem *showVDPPalette;
    IBOutlet NSMenuItem *showVDPPatterns;
    IBOutlet NSMenuItem *editCheats;
    IBOutlet NSMenuItem *enableCheats;

    IBOutlet NSMenuItem *showNESPatterns;

    IBOutlet NSPanel *vdpPalettePanel;
    IBOutlet PaletteGLView *paletteView;

    IBOutlet NSPanel *vdpPatternPanel;
    IBOutlet PatternGLView *patternView;

    IBOutlet NSPanel *nesPatternPanel;
    IBOutlet NESPatternGLView *nesPatternView;

    IBOutlet NSPanel *parPanel;

    IBOutlet SMSCheatController *cheatController;
    IBOutlet CrabEmuPrefsController *prefsController;

    IBOutlet NSWindow *sdscWindow;
    IBOutlet SDSCView *sdscView;

    IBOutlet NSMenuItem *normalRenderer;
    IBOutlet NSMenuItem *hq2xRenderer;
    IBOutlet NSMenuItem *hq4xRenderer;
    IBOutlet NSMenuItem *scale2xRenderer;

    NSMenuItem *p1curinput;
    NSMenuItem *p2curinput;

    NSUserDefaults *prefs;
    NSString *openRom;
    NSString *openRomSaveFn;
    NSString *cheatFn;
    NSMutableDictionary *keymap;

    int joy_count;

    gamepad_t *joy1;
    NSMutableDictionary *j1map;

    gamepad_t *joy2;
    NSMutableDictionary *j2map;

    NSTimer *timer;

    BOOL domestic;
    BOOL running;
    NSLock *runningLock;
}

/* NSApplication delegate methods */
- (void)application:(NSApplication *)app openFiles:(NSArray *)filenames;
- (NSApplicationTerminateReply)applicationShouldTerminate:(NSApplication *)app;

/* NSWindow delegate methods */
- (BOOL)windowShouldClose:(id)sender;

- (void)awakeFromNib;
- (void)dealloc;

- (IBAction)enableNoise:(id)sender;
- (IBAction)enablePSG:(id)sender;
- (IBAction)enableYM2413:(id)sender;
- (IBAction)enableTone0:(id)sender;
- (IBAction)enableTone1:(id)sender;
- (IBAction)enableTone2:(id)sender;
- (IBAction)loadMasterRom:(id)sender;
- (IBAction)enableDebugger:(id)sender;
- (IBAction)saveSRAM:(id)sender;
- (IBAction)loadSaveState:(id)sender;
- (IBAction)resetEmulation:(id)sender;
- (IBAction)regionDomestic:(id)sender;
- (IBAction)regionExport:(id)sender;
- (IBAction)modeNTSC:(id)sender;
- (IBAction)modePAL:(id)sender;
- (IBAction)setInput1:(id)sender;
- (IBAction)setInput2:(id)sender;
- (IBAction)takeScreenshot:(id)sender;
- (IBAction)toggleFullscreen:(id)sender;
- (IBAction)showVDPPalette:(id)sender;
- (IBAction)showVDPPatterns:(id)sender;
- (IBAction)toggleCheats:(id)sender;
- (IBAction)editCheats:(id)sender;
- (IBAction)showSDSCTerminal:(id)sender;
- (IBAction)changeRenderer:(id)sender;

- (IBAction)showNESPatterns:(id)sender;

- (void)eventFromGamepad:(gamepad_t *)pad element:(gp_elemdata_t *)elem
                   value:(int)value;
- (void)updatePalette;
- (void)updatePatterns;
- (void)updateSDSCTerminalWithString:(char[25][80 * 2])console;

- (void)masterEmuThread:(id)junk;
- (void)pauseEmulation:(id)sender;

- (NSString *)keyListForConsole:(int)console
                         player:(int)player;

- (BOOL)initMasterSystem:(int)console;
- (BOOL)loadMasterSystemRomFromFile:(NSString *)filename console:(int)c;

- (BOOL)initColecoVision;
- (BOOL)loadColecoVisionRomFromFile:(NSString *)filename;

- (BOOL)initNES;
- (BOOL)loadNESRomFromFile:(NSString *)filename;

- (BOOL)openRomFromFile:(NSString *)fn;

- (void)toggleScreensaver:(BOOL)on;

- (void)savePrefs;
- (void)readPrefs;

- (void)ssTimer:(NSTimer *)timer;

- (BOOL)running;

@end

#endif
