/*
    This file is part of CrabEmu.

    Copyright (C) 2009 Lawrence Sebald

    CrabEmu is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2 
    as published by the Free Software Foundation.

    CrabEmu is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CrabEmu; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <OpenGL/GL.h>
#include "PatternGLView.h"

#include "smsvdp.h"

/* From sms.c */
extern int sms_initialized;

@implementation PatternGLView

- (id)initWithFrame:(NSRect)frameRect
{
    NSOpenGLPixelFormatAttribute attrs[] = {
        NSOpenGLPFAWindow,
        NSOpenGLPFANoRecovery,
        NSOpenGLPFAColorSize, 32,
        NSOpenGLPFADepthSize, 32,
        NSOpenGLPFADoubleBuffer,
        0
    };

    NSOpenGLPixelFormat *fmt =
        [[NSOpenGLPixelFormat alloc] initWithAttributes:attrs];

    if(fmt == nil)
        return nil;

    self = [super initWithFrame:frameRect pixelFormat:fmt];

    [fmt release];

    return self;
}

- (void)drawRect:(NSRect)rect
{
    static BOOL first = YES;
    static GLuint tex;
    int i, j, k, x = 0, y = 0, tmp;
    uint16 buf[256 * 256] = { 0 };

    if(first) {
        glEnable(GL_TEXTURE_2D);
        glGenTextures(1, &tex);
        glBindTexture(GL_TEXTURE_2D, tex);
        glTexImage2D(GL_TEXTURE_2D, 0, 4, 256, 256, 0, GL_BGRA,
                     GL_UNSIGNED_SHORT_1_5_5_5_REV, buf);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
        glDisable(GL_TEXTURE_2D);

        glDisable(GL_BLEND);
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

        glShadeModel(GL_SMOOTH);
        glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
        glClearDepth(0.0f);

        glViewport(0.0f, 0.0f, [self bounds].size.width,
                   [self bounds].size.height);

        glMatrixMode(GL_PROJECTION);
        glLoadIdentity();

        glOrtho(0.0f, 256.0f, 256.0f, 0.0f, 1.0f, 0.0f);

        glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();

        first = NO;
    }

    if(sms_initialized) {
        for(i = 0; i < 512; ++i) {
            if(smsvdp.pattern[i].dirty)
                sms_vdp_update_cache(i);

            for(j = 0; j < 8; ++j) {
                for(k = 0; k < 8; ++k) {
                    /* Draw with background palette first. */
                    tmp = smsvdp.pattern[i].texture[0][j * 8 + k];
                    buf[(y + j) * 256 + x + k] = smsvdp.pal[tmp];

                    /* Then with the sprite palette. */
                    tmp |= 0x10;
                    buf[(y + j + 128) * 256 + x + k] = smsvdp.pal[tmp];
                    
                }
            }

            x += 8;

            if(x == 256) {
                x = 0;
                y += 8;
            }
        }

        glClear(GL_COLOR_BUFFER_BIT);
        glLoadIdentity();

        glEnable(GL_TEXTURE_2D);

        glBindTexture(GL_TEXTURE_2D, tex);
        glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, 256, 256, GL_BGRA,
                        GL_UNSIGNED_SHORT_1_5_5_5_REV, buf);

        glBegin(GL_QUADS);
        glTexCoord2f(0.0f, 0.0f);
        glVertex2f(0.0f, 0.0f);
        glTexCoord2f(1.0f, 0.0f);
        glVertex2f(256.0f, 0.0f);
        glTexCoord2f(1.0f, 1.0f);
        glVertex2f(256.0f, 256.0f);
        glTexCoord2f(0.0f, 1.0f);
        glVertex2f(0.0f, 256.0f);
        glEnd();

        glDisable(GL_TEXTURE_2D);
    }

    glFlush();
    [[self openGLContext] flushBuffer];
}

@end
