/*
    This file is part of CrabEmu.

    Copyright (C) 2009 Lawrence Sebald

    CrabEmu is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2 
    as published by the Free Software Foundation.

    CrabEmu is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CrabEmu; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <OpenGL/GL.h>
#include "SDSCView.h"
#include "OpenGLFont.h"

#include "sdscterminal.h"

extern int sms_initialized;

/* Color table for the SDSC attribute byte. */
static float attr_lut[16 * 4] = {
    0.0f,   0.0f,   0.0f,   1.0f,   /* Black */
    0.0f,   0.0f,   0.25f,  1.0f,   /* Dark Blue */
    0.0f,   0.25f,  0.0f,   1.0f,   /* Dark Green */
    0.0f,   0.25f,  0.25f,  1.0f,   /* Dark Cyan */
    0.25f,  0.0f,   0.0f,   1.0f,   /* Dark Red */
    0.5f,   0.0f,   0.25f,  1.0f,   /* Dark Magenta */
    0.5f,   0.5f,   0.0f,   1.0f,   /* Dark Yellow */
    0.25f,  0.25f,  0.25f,  1.0f,   /* Dark Gray */
    0.75f,  0.75f,  0.75f,  1.0f,   /* Light Gray */
    0.0f,   0.0f,   1.0f,   1.0f,   /* Light Blue */
    0.0f,   1.0f,   0.0f,   1.0f,   /* Light Green */
    0.0f,   1.0f,   1.0f,   1.0f,   /* Light Cyan */
    1.0f,   0.0f,   0.0f,   1.0f,   /* Light Red */
    1.0f,   0.0f,   0.5f,   1.0f,   /* Light Magenta */
    1.0f,   1.0f,   0.0f,   1.0f,   /* Light Yellow */
    1.0f,   1.0f,   1.0f,   1.0f    /* White */
};

@implementation SDSCView

- (id)initWithFrame:(NSRect)frameRect
{
    NSOpenGLPixelFormatAttribute attrs[] = {
        NSOpenGLPFAWindow,
        NSOpenGLPFANoRecovery,
        NSOpenGLPFAColorSize, 32,
        NSOpenGLPFADepthSize, 32,
        NSOpenGLPFADoubleBuffer,
        0
    };

    NSOpenGLPixelFormat *fmt =
        [[NSOpenGLPixelFormat alloc] initWithAttributes:attrs];

    if(fmt == nil)
        return nil;

    if((self = [super initWithFrame:frameRect pixelFormat:fmt]) == nil) {
        [fmt release];
        [font release];
        return nil;
    }

    [fmt release];

    [self resetConsole];

    return self;
}

- (void)dealloc
{
    [font release];
    [super dealloc];
}

- (void)drawRect:(NSRect)rect
{
    static BOOL first = YES;

    if(first) {
        NSFont *f = [NSFont fontWithName:@"Monaco" size:10.0];

        glShadeModel(GL_SMOOTH);
        glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
        glClearDepth(1.0f);

        glEnable(GL_BLEND);
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

        glViewport(0.0f, 0.0f, [self bounds].size.width,
                   [self bounds].size.height);

        glMatrixMode(GL_PROJECTION);
        glLoadIdentity();

        glOrtho(0.0f, 480.0f, 350.0f, 0.0f, 1.0f, 0.0f);

        glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();

        font = [[OpenGLFont alloc] initWithFont:f];

        /* We're using a monospaced font, so this will do just fine. */
        charHeight = [font metricsForCharacter:'A'].height;

        first = NO;
    }

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    if(consoleData != NULL) {
        int i, j, fg, bg;
        NSPoint pt = NSMakePoint(0.0, 0.0);

        for(i = 0; i < 25; ++i) {
            for(j = 0; j < 80; ++j) {
                fg = (consoleData[i][j << 1] & 0x0F) << 2;
                bg = ((consoleData[i][j << 1] >> 4) & 0x0F) << 2;

                pt = [font drawCharacter:consoleData[i][(j << 1) + 1]
                                      at:pt
                                     red:attr_lut[fg]
                                   green:attr_lut[fg + 1]
                                    blue:attr_lut[fg + 2]
                                   alpha:attr_lut[fg + 3]
                                   bgRed:attr_lut[bg]
                                 bgGreen:attr_lut[bg + 1]
                                  bgBlue:attr_lut[bg + 2]
                                 bgAlpha:attr_lut[bg + 3]];
            }

            pt.x = 0.0;
            pt.y += charHeight;
        }
    }

    glFlush();
    [[self openGLContext] flushBuffer];
}

- (void)updateString:(char[25][80 * 2])console
{
    int i;

    for(i = 0; i < 25; ++i) {
        memcpy(consoleData[i], console[i], 80 * 2);
    }

    [self setNeedsDisplay:YES];
}

- (void)resetConsole
{
    int i, j;

    for(i = 0; i < 25; ++i) {
        for(j = 0; j < 80; ++j) {
            consoleData[i][j << 1] = 0;
            consoleData[i][(j << 1) + 1] = ' ';
        }
    }
}

@end
