/*
    This file is part of CrabEmu.

    Copyright (C) 2009 Lawrence Sebald

    CrabEmu is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2 
    as published by the Free Software Foundation.

    CrabEmu is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CrabEmu; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "SMSCheatController.h"

#include "list.h"
#include "cheats.h"

extern cc_dlist_t *sms_active_cheats;
extern NSLock *memoryLock;

@implementation SMSCheatController

- (IBAction)addCheat:(id)sender
{
    int rv;

    [NSApp beginSheet:codePanel
       modalForWindow:cheatPanel
        modalDelegate:self
       didEndSelector:nil
          contextInfo:nil];

    for(;;) {
        rv = [NSApp runModalForWindow:codePanel];

        if(rv == NSOKButton) {
            NSString *c1, *c2, *d;
            int tmp;
            uint32 code;
            const char *str;
            sms_cheat_t *cht;

            c1 = [code1 stringValue];

            if([c1 length] != 4) {
                NSRunAlertPanel(@"Invalid code", @"The code entered is not of "
                                "the correct length.", @"OK", nil, nil);
                continue;
            }

            str = [c1 UTF8String];
            if(sscanf(str, "%04X", &tmp) != 1) {
                NSRunAlertPanel(@"Invalid code", @"The code entered is not of "
                                "the correct format.", @"OK", nil, nil);
                continue;
            }

            code = tmp << 16;

            c2 = [code2 stringValue];

            if([c2 length] != 4) {
                NSRunAlertPanel(@"Invalid code", @"The code entered is not of "
                                "the correct length.", @"OK", nil, nil);
                continue;
            }

            str = [c2 UTF8String];
            if(sscanf(str, "%04X", &tmp) != 1) {
                NSRunAlertPanel(@"Invalid code", @"The code entered is not of "
                                "the correct format.", @"OK", nil, nil);
                continue;
            }
            
            code |= tmp;

            d = [description stringValue];

            if([d length] > 63) {
                NSRunAlertPanel(@"Invalid description", @"The description is "
                                "too long.", @"OK", nil, nil);
                continue;
            }

            cht = (sms_cheat_t *)malloc(sizeof(sms_cheat_t));
            if(cht) {
                cht->ar_code = code;
                strcpy(cht->desc, [d UTF8String]);
                cht->enabled = 1;

                [memoryLock lock];
                sms_cheat_add(cht);
                [memoryLock unlock];
            }

            break;
        }
        else {
            break;
        }
    }

    [code1 setStringValue:@""];
    [code2 setStringValue:@""];
    [description setStringValue:@""];

    [NSApp endSheet:codePanel];
    [codePanel orderOut:self];

    [self fillTable];
}

- (IBAction)deleteCheat:(id)sender
{
    int row = [cheatView selectedRow];

    if(row == -1)
        return;

    [memoryLock lock];

    cc_dlist_remove_at(sms_active_cheats, (uint32)row);

    [memoryLock unlock];
    [self fillTable];
}

- (IBAction)cheatOK:(id)sender
{
    [NSApp stopModalWithCode:NSOKButton];
}

- (IBAction)cheatCancel:(id)sender
{
    [NSApp stopModalWithCode:NSCancelButton];
}

- (void)fillTable
{
    [cheatView reloadData];
}

- (int)numberOfRowsInTableView:(NSTableView *)v
{
    if(sms_active_cheats)
        return sms_active_cheats->count;
    else
        return 0;
}

-           (id)tableView:(NSTableView *)v
objectValueForTableColumn:(NSTableColumn *)tc
                      row:(int)idx
{
    NSString *ident = (NSString *)[tc identifier];
    sms_cheat_t *c =
        (sms_cheat_t *)cc_dlist_get_data_at(sms_active_cheats, idx);

    if([ident isEqualToString:@"Code"]) {
        return [NSString stringWithFormat:@"%04X-%04X", (c->ar_code >> 16),
            (c->ar_code & 0xFFFF)];
    }
    else {
        return [NSString stringWithUTF8String:c->desc];
    }
}

- (void)tableView:(NSTableView *)v
   setObjectValue:(id)o
   forTableColumn:(NSTableColumn *)tc
              row:(int)idx
{
    NSString *ident = (NSString *)[tc identifier];
    sms_cheat_t *c =
        (sms_cheat_t *)cc_dlist_get_data_at(sms_active_cheats, idx);
    NSString *s = (NSString *)o;

    if([ident isEqualToString:@"Description"]) {
        if([s length] > 63) {
            NSRunAlertPanel(@"Invalid description", @"The description is "
                            "too long.", @"OK", nil, nil);
            return;
        }

        strcpy(c->desc, [s UTF8String]);
    }
    else if([ident isEqualToString:@"Code"]) {
        uint32 num, num2;

        if([s length] != 9) {
            NSRunAlertPanel(@"Invalid Code", @"The code is not of the correct "
                            "format.", @"OK", nil, nil);
            return;
        }

        if(sscanf([s UTF8String], "%04X-%04X", &num, &num2) != 2) {
            NSRunAlertPanel(@"Invalid Code", @"The code is not of the correct "
                            "format.", @"OK", nil, nil);
            return;
        }

        c->ar_code = (num << 16) | num2;
    }
}

@end
