/*
    This file is part of CrabEmu.

    Copyright (C) 2006 Lawrence Sebald

    CrabEmu is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2 
    as published by the Free Software Foundation.

    CrabEmu is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CrabEmu; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#import "Z80DebugController.h"
#include "NSCondition.h"

#include "../cpu/CrabZ80/CrabZ80.h"
#include "../cpu/CrabZ80/CrabZ80d.h"

extern CrabZ80_t *cpuz80;
extern Z80DebugController *debugger;
extern int debuggerMode;
extern NSCondition *ss_sem;

@implementation Z80DebugController

- (void)awakeFromNib
{
    debugger = self;
}

- (void)updateWithCycles:(int)cycles scanline:(int)scanline
{
    char string[64];

    CrabZ80_disassemble(string, cpuz80, cpuz80->pc.w);
    [currentInstruction setStringValue: [NSString stringWithUTF8String: string]];

    sprintf(string, "%d", cycles);
    [cyclesToGo setStringValue: [NSString stringWithUTF8String: string]];

    sprintf(string, "%d", scanline);
    [scanlineNumber setStringValue: [NSString stringWithUTF8String: string]];

    sprintf(string, "%04X", cpuz80->af.w);
    [regAF setStringValue: [NSString stringWithUTF8String: string]];

    sprintf(string, "%04X", cpuz80->bc.w);
    [regBC setStringValue: [NSString stringWithUTF8String: string]];

    sprintf(string, "%04X", cpuz80->de.w);
    [regDE setStringValue: [NSString stringWithUTF8String: string]];

    sprintf(string, "%04X", cpuz80->hl.w);
    [regHL setStringValue: [NSString stringWithUTF8String: string]];

    sprintf(string, "%04X", cpuz80->ix.w);
    [regIX setStringValue: [NSString stringWithUTF8String: string]];

    sprintf(string, "%04X", cpuz80->iy.w);
    [regIY setStringValue: [NSString stringWithUTF8String: string]];

    sprintf(string, "%04X", cpuz80->pc.w);
    [regPC setStringValue: [NSString stringWithUTF8String: string]];

    sprintf(string, "%04X", cpuz80->sp.w);
    [regSP setStringValue: [NSString stringWithUTF8String: string]];

    sprintf(string, "%04X", cpuz80->ir.w);
    [regIR setStringValue: [NSString stringWithUTF8String: string]];

    sprintf(string, "%04X", cpuz80->afp.w);
    [regAFp setStringValue: [NSString stringWithUTF8String: string]];

    sprintf(string, "%04X", cpuz80->bcp.w);
    [regBCp setStringValue: [NSString stringWithUTF8String: string]];

    sprintf(string, "%04X", cpuz80->dep.w);
    [regDEp setStringValue: [NSString stringWithUTF8String: string]];

    sprintf(string, "%04X", cpuz80->hlp.w);
    [regHLp setStringValue: [NSString stringWithUTF8String: string]];

    if(cpuz80->iff1)
        [stateIFF1 setState: NSOnState];
    else
        [stateIFF1 setState: NSOffState];

    if(cpuz80->iff2)
        [stateIFF2 setState: NSOnState];
    else
        [stateIFF2 setState: NSOffState];

    if(cpuz80->halt)
        [stateHalted setState: NSOnState];
    else
        [stateHalted setState: NSOffState];

    if(CRABZ80_GET_FLAG(cpuz80, CRABZ80_CF))
        [flagsC setState: NSOnState];
    else
        [flagsC setState: NSOffState];

    if(CRABZ80_GET_FLAG(cpuz80, CRABZ80_NF))
        [flagsN setState: NSOnState];
    else
        [flagsN setState: NSOffState];

    if(CRABZ80_GET_FLAG(cpuz80, CRABZ80_PF))
        [flagsP setState: NSOnState];
    else
        [flagsP setState: NSOffState];

    if(CRABZ80_GET_FLAG(cpuz80, CRABZ80_XF))
        [flagsX setState: NSOnState];
    else
        [flagsX setState: NSOffState];

    if(CRABZ80_GET_FLAG(cpuz80, CRABZ80_HF))
        [flagsH setState: NSOnState];
    else
        [flagsH setState: NSOffState];

    if(CRABZ80_GET_FLAG(cpuz80, CRABZ80_YF))
        [flagsY setState: NSOnState];
    else
        [flagsY setState: NSOffState];

    if(CRABZ80_GET_FLAG(cpuz80, CRABZ80_ZF))
        [flagsZ setState: NSOnState];
    else
        [flagsZ setState: NSOffState];

    if(CRABZ80_GET_FLAG(cpuz80, CRABZ80_SF))
        [flagsS setState: NSOnState];
    else
        [flagsS setState: NSOffState];
}

-(IBAction)singleStep:(id)sender
{
    debuggerMode = 0;
    [ss_sem signal];
}

- (IBAction)runScanline:(id)sender
{
    debuggerMode = 1;
    [ss_sem signal];
}

- (IBAction)runFrame:(id)sender
{
    debuggerMode = 2;
    [ss_sem signal];
}

@end
