/*
    This file is part of CrabEmu.

    Copyright (C) 2009, 2013 Lawrence Sebald

    CrabEmu is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2
    as published by the Free Software Foundation.

    CrabEmu is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CrabEmu; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <QMutex>
#include "CrabEmuThread.h"
#include "sound.h"

extern QMutex memoryLock;
extern int emu_frame(int run);
extern void render_scene(void);
extern double GetAbsoluteTime(void);
extern "C" int sms_initialized;
extern "C" int coleco_initialized;
extern "C" int nes_initialized;

CrabEmuThread::CrabEmuThread(QObject *parent) : QThread(parent) {
    finished = false;
    fps = 60.0;
}

CrabEmuThread::~CrabEmuThread() {
    finish();
    wait();
}

void CrabEmuThread::restart() {
    finished = false;
    start();
}

void CrabEmuThread::run() {
    int run = 0;

    while(!finished) {
        memoryLock.lock();
        if(!sms_initialized && !coleco_initialized && !nes_initialized) {
            emit frameRendered();
            memoryLock.unlock();
            sleep(1);
            continue;
        }

        run = emu_frame(run);

        render_scene();
        emit frameRendered();
        memoryLock.unlock();

        sound_wait();
    }
}

void CrabEmuThread::finish() {
    finished = 1;
}

void CrabEmuThread::setFps(float f) {
    fps = f;
}
