/*
    This file is part of CrabEmu.

    Copyright (C) 2009, 2013 Lawrence Sebald

    CrabEmu is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2
    as published by the Free Software Foundation.

    CrabEmu is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CrabEmu; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#ifndef EMUWIDGET_H
#define EMUWIDGET_H

#include <QWidget>
#include <QImage>
#include <QPixmap>
#include <QMutex>

#ifdef ENABLE_OPENGL
#include <QGLWidget>
#endif

#include "CrabEmu.h"

class CrabEmuThread;

class QMouseEvent;

typedef enum {
    Render_Normal,
    Render_Scale2x,
    Render_HQ2x,
    Render_HQ4x
} EmuRenderer;

#ifndef ENABLE_OPENGL

class EmuWidget : public QWidget {
        Q_OBJECT

    public:
        EmuWidget(CrabEmuThread *thd, QWidget *parent = NULL);

        QImage *retainImage();
        void releaseImage();
        void setRenderer(EmuRenderer r);

    protected:
        void paintEvent(QPaintEvent *event);
        void mousePressEvent(QMouseEvent *event);
        void mouseReleaseEvent(QMouseEvent *event);
        void mouseMoveEvent(QMouseEvent *event);

    private:
        QMutex mutex;
        QImage img;
        EmuRenderer renderer;
};

#else

class EmuWidget : public QGLWidget {
    Q_OBJECT

    public:
        EmuWidget(CrabEmuThread *thd, QWidget *parent = NULL);

        QImage *retainImage();
        void releaseImage();
        void setRenderer(EmuRenderer r);

    protected:
        void resizeGL(int width, int height);
        void paintGL();
        void mousePressEvent(QMouseEvent *event);
        void mouseReleaseEvent(QMouseEvent *event);
        void mouseMoveEvent(QMouseEvent *event);

    private:
        QImage img;
        QMutex mutex;
        GLuint glTexture;
        bool txrCreated;
        bool txrChanged;
        int txrWidth;
        int txrHeight;
        int cxtWidth;
        int cxtHeight;
        EmuRenderer renderer;
        pixel_t *fbTrans;
};

#endif /* ENABLE_OPENGL */

#endif /* EMUWIDGET_H */
