/*
    This file is part of CrabEmu.

    Copyright (C) 2009, 2013 Lawrence Sebald

    CrabEmu is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2
    as published by the Free Software Foundation.

    CrabEmu is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CrabEmu; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#ifndef UICRABEMU_H
#define UICRABEMU_H

#include "ui_CrabEmu.h"
#include "CrabEmuThread.h"

class EmuWidget;

class UIDebugWindow;

class UICrabEmu : public QMainWindow, public Ui::CrabEmu {
        Q_OBJECT

    public:
        UICrabEmu(QWidget *parent = NULL);
        virtual ~UICrabEmu();
        EmuWidget *emuwidget;
        CrabEmuThread *thread();
        UIDebugWindow *debug();
        void loadRom(QString file);
        void restoreSettings();

    protected:
        void closeEvent(QCloseEvent *event);
        void keyPressEvent(QKeyEvent *e);
        void keyReleaseEvent(QKeyEvent *e);
        void saveLoad(QString state, bool load);
        void loadSRAM();
        void saveSRAM();

    signals:
        void pause(bool relay);

    public slots:
        void on_actionPause_triggered();
        void on_actionDebugger_triggered();
        void on_actionReset_triggered();
        void on_actionReloadRom_triggered();
        void on_Debugger_destroyed();

    protected slots:
        void on_actionOpenRom_triggered();
        void on_actionSaveSRAM_triggered();

        void on_actionSaveSlot1_triggered();
        void on_actionSaveSlot2_triggered();
        void on_actionSaveSlot3_triggered();
        void on_actionSaveSlot4_triggered();
        void on_actionSaveSlot5_triggered();
        void on_actionSaveSlot6_triggered();

        void on_actionLoadSlot1_triggered();
        void on_actionLoadSlot2_triggered();
        void on_actionLoadSlot3_triggered();
        void on_actionLoadSlot4_triggered();
        void on_actionLoadSlot5_triggered();
        void on_actionLoadSlot6_triggered();

        void on_actionTakeScreenshot_triggered();

        void on_actionExit_triggered();

        void on_actionEnablePSG_triggered();
        void on_actionEnableYM2413_triggered();
        void on_actionTone0_triggered();
        void on_actionTone1_triggered();
        void on_actionTone2_triggered();
        void on_actionNoise_triggered();

        void on_actionExport_triggered();
        void on_actionDomestic_triggered();

        void on_actionNTSC_triggered();
        void on_actionPAL_triggered();

        void on_actionEnableCheats_triggered();

        void on_actionNormal_triggered();
        void on_actionScale2x_triggered();
        void on_actionHQ2x_triggered();
        void on_actionHQ4x_triggered();

        void on_actionShutdown_triggered();

        void on_actionPreferences_triggered();

        void on_actionAbout_triggered();

    private:
        QString openRom;
        QString romName;
        QString romExt;
        CrabEmuThread *thd;
        bool paused;
        bool debugger;
        UIDebugWindow *debugWindow;
        int region;
        int videoSystem;
        int videoMultiple;

        int initializeSMS(QString filename, int console);
        int initializeCV(QString filename);
        int initializeNES(QString filename);
        void initializeSettings();
        void resizeEmu(unsigned w, unsigned h);
        void unpause(bool p = true);
        void updateFps();
        void smsKeyPressEvent(QKeyEvent *e);
        void smsKeyReleaseEvent(QKeyEvent *e);
        void colecoKeyPressEvent(QKeyEvent *e);
        void colecoKeyReleaseEvent(QKeyEvent *e);
        void nesKeyPressEvent(QKeyEvent *e);
        void nesKeyReleaseEvent(QKeyEvent *e);
};

#endif /* !UICRABEMU_H */
