/*
    This file is part of CrabEmu.

    Copyright (C) 2009, 2013 Lawrence Sebald

    CrabEmu is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2
    as published by the Free Software Foundation.

    CrabEmu is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CrabEmu; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <QFileDialog>
#include <QSettings>

#include "UIPreferences.h"
#include "UIKeyInputDialog.h"
#include "CrabEmu.h"
#include "sms.h"
#include "colecovision.h"
#include "nes.h"

UIPreferences::UIPreferences(QWidget *parent) : QDialog(parent) {
    QSettings settings;

    setupUi(this);

    sms_bm = settings.value("SMS Buttonmap").toMap();
    coleco_bm = settings.value("Coleco Buttonmap").toMap();
    nes_bm = settings.value("NES Buttonmap").toMap();
    editSMSBios->setText(settings.value("SMS BIOS ROM").toString());
    editGGBios->setText(settings.value("Game Gear BIOS ROM").toString());
    editCVBios->setText(settings.value("ColecoVision BIOS ROM").toString());
    smsBiosEnabled->setChecked(settings.value("SMS BIOS Enabled").toBool());
    ggBiosEnabled->setChecked(settings.value("GG BIOS Enabled").toBool());

    editSave->setText(settings.value("Save Directory").toString());
    editState->setText(settings.value("State Directory").toString());
    editCheat->setText(settings.value("Cheat Directory").toString());

    saveEnabled->setChecked(settings.value("Save Directory Enabled").toBool());
    stateEnabled->setChecked(settings.value("State Directory Enabled").toBool());
    cheatEnabled->setChecked(settings.value("Cheat Directory Enabled").toBool());
}

void UIPreferences::updatePreferences() {
    QSettings settings;

    settings.setValue("SMS Buttonmap", sms_bm);
    settings.setValue("Coleco Buttonmap", coleco_bm);
    settings.setValue("SMS BIOS ROM", editSMSBios->text());
    settings.setValue("Game Gear BIOS ROM", editGGBios->text());
    settings.setValue("ColecoVision BIOS ROM", editCVBios->text());
    settings.setValue("SMS BIOS Enabled", smsBiosEnabled->isChecked());
    settings.setValue("GG BIOS Enabled", ggBiosEnabled->isChecked());

    settings.setValue("Save Directory", editSave->text());
    settings.setValue("State Directory", editState->text());
    settings.setValue("Cheat Directory", editState->text());

    settings.setValue("Save Directory Enabled", saveEnabled->isChecked());
    settings.setValue("State Directory Enabled", stateEnabled->isChecked());
    settings.setValue("Cheat Directory Enabled", cheatEnabled->isChecked());
}

void UIPreferences::on_buttonSMS1Up_clicked() {
    setKey("Qt Keyboard 1", "SMS Up");
}

void UIPreferences::on_buttonSMS1Down_clicked() {
    setKey("Qt Keyboard 1", "SMS Down");
}

void UIPreferences::on_buttonSMS1Left_clicked() {
    setKey("Qt Keyboard 1", "SMS Left");
}

void UIPreferences::on_buttonSMS1Right_clicked() {
    setKey("Qt Keyboard 1", "SMS Right");
}

void UIPreferences::on_buttonSMS1A_clicked() {
    setKey("Qt Keyboard 1", "SMS Button 1");
}

void UIPreferences::on_buttonSMS1B_clicked() {
    setKey("Qt Keyboard 1", "SMS Button 2");
}

void UIPreferences::on_buttonSMS2Up_clicked() {
    setKey("Qt Keyboard 2", "SMS Up");
}

void UIPreferences::on_buttonSMS2Down_clicked() {
    setKey("Qt Keyboard 2", "SMS Down");
}

void UIPreferences::on_buttonSMS2Left_clicked() {
    setKey("Qt Keyboard 2", "SMS Left");
}

void UIPreferences::on_buttonSMS2Right_clicked() {
    setKey("Qt Keyboard 2", "SMS Right");
}

void UIPreferences::on_buttonSMS2A_clicked() {
    setKey("Qt Keyboard 2", "SMS Button 1");
}

void UIPreferences::on_buttonSMS2B_clicked() {
    setKey("Qt Keyboard 2", "SMS Button 2");
}

void UIPreferences::on_buttonGGStart_clicked() {
    setKey("Qt Keyboard 1", "Game Gear Start");
}

void UIPreferences::on_buttonColeco1Up_clicked() {
    setKey("Qt Keyboard 1", "Coleco Up", CONSOLE_COLECOVISION);
}

void UIPreferences::on_buttonColeco1Down_clicked() {
    setKey("Qt Keyboard 1", "Coleco Down", CONSOLE_COLECOVISION);
}

void UIPreferences::on_buttonColeco1Left_clicked() {
    setKey("Qt Keyboard 1", "Coleco Left", CONSOLE_COLECOVISION);
}

void UIPreferences::on_buttonColeco1Right_clicked() {
    setKey("Qt Keyboard 1", "Coleco Right", CONSOLE_COLECOVISION);
}

void UIPreferences::on_buttonColeco1LAction_clicked() {
    setKey("Qt Keyboard 1", "Coleco L Action", CONSOLE_COLECOVISION);
}

void UIPreferences::on_buttonColeco1RAction_clicked() {
    setKey("Qt Keyboard 1", "Coleco R Action", CONSOLE_COLECOVISION);
}

void UIPreferences::on_buttonColeco1one_clicked() {
    setKey("Qt Keyboard 1", "Coleco 1", CONSOLE_COLECOVISION);
}

void UIPreferences::on_buttonColeco1two_clicked() {
    setKey("Qt Keyboard 1", "Coleco 2", CONSOLE_COLECOVISION);
}

void UIPreferences::on_buttonColeco1three_clicked() {
    setKey("Qt Keyboard 1", "Coleco 3", CONSOLE_COLECOVISION);
}

void UIPreferences::on_buttonColeco1four_clicked() {
    setKey("Qt Keyboard 1", "Coleco 4", CONSOLE_COLECOVISION);
}

void UIPreferences::on_buttonColeco1five_clicked() {
    setKey("Qt Keyboard 1", "Coleco 5", CONSOLE_COLECOVISION);
}

void UIPreferences::on_buttonColeco1six_clicked() {
    setKey("Qt Keyboard 1", "Coleco 6", CONSOLE_COLECOVISION);
}

void UIPreferences::on_buttonColeco1seven_clicked() {
    setKey("Qt Keyboard 1", "Coleco 7", CONSOLE_COLECOVISION);
}

void UIPreferences::on_buttonColeco1eight_clicked() {
    setKey("Qt Keyboard 1", "Coleco 8", CONSOLE_COLECOVISION);
}

void UIPreferences::on_buttonColeco1nine_clicked() {
    setKey("Qt Keyboard 1", "Coleco 9", CONSOLE_COLECOVISION);
}

void UIPreferences::on_buttonColeco1star_clicked() {
    setKey("Qt Keyboard 1", "Coleco *", CONSOLE_COLECOVISION);
}

void UIPreferences::on_buttonColeco1zero_clicked() {
    setKey("Qt Keyboard 1", "Coleco 0", CONSOLE_COLECOVISION);
}

void UIPreferences::on_buttonColeco1pound_clicked() {
    setKey("Qt Keyboard 1", "Coleco #", CONSOLE_COLECOVISION);
}

void UIPreferences::on_buttonColeco2Up_clicked() {
    setKey("Qt Keyboard 2", "Coleco Up", CONSOLE_COLECOVISION);
}

void UIPreferences::on_buttonColeco2Down_clicked() {
    setKey("Qt Keyboard 2", "Coleco Down", CONSOLE_COLECOVISION);
}

void UIPreferences::on_buttonColeco2Left_clicked() {
    setKey("Qt Keyboard 2", "Coleco Left", CONSOLE_COLECOVISION);
}

void UIPreferences::on_buttonColeco2Right_clicked() {
    setKey("Qt Keyboard 2", "Coleco Right", CONSOLE_COLECOVISION);
}

void UIPreferences::on_buttonColeco2LAction_clicked() {
    setKey("Qt Keyboard 2", "Coleco L Action", CONSOLE_COLECOVISION);
}

void UIPreferences::on_buttonColeco2RAction_clicked() {
    setKey("Qt Keyboard 2", "Coleco R Action", CONSOLE_COLECOVISION);
}

void UIPreferences::on_buttonColeco2one_clicked() {
    setKey("Qt Keyboard 2", "Coleco 1", CONSOLE_COLECOVISION);
}

void UIPreferences::on_buttonColeco2two_clicked() {
    setKey("Qt Keyboard 2", "Coleco 2", CONSOLE_COLECOVISION);
}

void UIPreferences::on_buttonColeco2three_clicked() {
    setKey("Qt Keyboard 2", "Coleco 3", CONSOLE_COLECOVISION);
}

void UIPreferences::on_buttonColeco2four_clicked() {
    setKey("Qt Keyboard 2", "Coleco 4", CONSOLE_COLECOVISION);
}

void UIPreferences::on_buttonColeco2five_clicked() {
    setKey("Qt Keyboard 2", "Coleco 5", CONSOLE_COLECOVISION);
}

void UIPreferences::on_buttonColeco2six_clicked() {
    setKey("Qt Keyboard 2", "Coleco 6", CONSOLE_COLECOVISION);
}

void UIPreferences::on_buttonColeco2seven_clicked() {
    setKey("Qt Keyboard 2", "Coleco 7", CONSOLE_COLECOVISION);
}

void UIPreferences::on_buttonColeco2eight_clicked() {
    setKey("Qt Keyboard 2", "Coleco 8", CONSOLE_COLECOVISION);
}

void UIPreferences::on_buttonColeco2nine_clicked() {
    setKey("Qt Keyboard 2", "Coleco 9", CONSOLE_COLECOVISION);
}

void UIPreferences::on_buttonColeco2star_clicked() {
    setKey("Qt Keyboard 2", "Coleco *", CONSOLE_COLECOVISION);
}

void UIPreferences::on_buttonColeco2zero_clicked() {
    setKey("Qt Keyboard 2", "Coleco 0", CONSOLE_COLECOVISION);
}

void UIPreferences::on_buttonColeco2pound_clicked() {
    setKey("Qt Keyboard 2", "Coleco #", CONSOLE_COLECOVISION);
}

void UIPreferences::on_buttonNES1Up_clicked() {
    setKey("Qt Keyboard 1", "Up", CONSOLE_NES);
}

void UIPreferences::on_buttonNES1Down_clicked() {
    setKey("Qt Keyboard 1", "Down", CONSOLE_NES);
}

void UIPreferences::on_buttonNES1Left_clicked() {
    setKey("Qt Keyboard 1", "Left", CONSOLE_NES);
}

void UIPreferences::on_buttonNES1Right_clicked() {
    setKey("Qt Keyboard 1", "Right", CONSOLE_NES);
}

void UIPreferences::on_buttonNES1A_clicked() {
    setKey("Qt Keyboard 1", "A", CONSOLE_NES);
}

void UIPreferences::on_buttonNES1B_clicked() {
    setKey("Qt Keyboard 1", "B", CONSOLE_NES);
}

void UIPreferences::on_buttonNES1Start_clicked() {
    setKey("Qt Keyboard 1", "Start", CONSOLE_NES);
}

void UIPreferences::on_buttonNES1Select_clicked() {
    setKey("Qt Keyboard 1", "Select", CONSOLE_NES);
}

void UIPreferences::on_buttonNES2Up_clicked() {
    setKey("Qt Keyboard 2", "Up", CONSOLE_NES);
}

void UIPreferences::on_buttonNES2Down_clicked() {
    setKey("Qt Keyboard 2", "Down", CONSOLE_NES);
}

void UIPreferences::on_buttonNES2Left_clicked() {
    setKey("Qt Keyboard 2", "Left", CONSOLE_NES);
}

void UIPreferences::on_buttonNES2Right_clicked() {
    setKey("Qt Keyboard 2", "Right", CONSOLE_NES);
}

void UIPreferences::on_buttonNES2A_clicked() {
    setKey("Qt Keyboard 2", "A", CONSOLE_NES);
}

void UIPreferences::on_buttonNES2B_clicked() {
    setKey("Qt Keyboard 2", "B", CONSOLE_NES);
}

void UIPreferences::on_buttonNES2Start_clicked() {
    setKey("Qt Keyboard 2", "Start", CONSOLE_NES);
}

void UIPreferences::on_buttonNES2Select_clicked() {
    setKey("Qt Keyboard 2", "Select", CONSOLE_NES);
}

void UIPreferences::on_buttonBrowseSMSBios_clicked() {
    openDialog("Select a SMS BIOS Rom", "SMS Roms (*.sms)", editSMSBios);
}

void UIPreferences::on_buttonBrowseGGBios_clicked() {
    openDialog("Select a Game Gear BIOS Rom", "Game Gear Roms (*.gg)",
            editGGBios);
}


void UIPreferences::on_buttonBrowseCVBios_clicked() {
    openDialog("Select a ColecoVision BIOS Rom", "ColecoVision Roms (*.rom *.col)", editCVBios);
}

void UIPreferences::on_buttonBrowseSave_clicked() {
    openDirectory("Select a Directory to Save", editSave);
}

void UIPreferences::on_buttonBrowseState_clicked() {
    openDirectory("Select a Directory to Save States", editState);
}

void UIPreferences::on_buttonBrowseCheat_clicked() {
    openDirectory("Select a Directory to Save Cheats", editCheat);
}

void UIPreferences::openDialog(const char *caption, const char *filter,
                QLineEdit *line) {
    QString fn;

    fn = QFileDialog::getOpenFileName(QApplication::activeWindow(),
                                tr(caption), "",
                                tr(filter),
                                // this is obsolete & does nothing since Qt 4.5
                                0, QFileDialog::DontUseSheet);

    /* If a file was not selected, then don't try to open one. */
    if(fn.isEmpty()) {
        return;
    }

    line->setText(fn);
}

void UIPreferences::openDirectory(const char *caption, QLineEdit *line) {
    QString fn;

    fn = QFileDialog::getExistingDirectory(QApplication::activeWindow(),
                                           tr(caption));

    /* if a directory was not selected, then don't do anything. */
    if(fn.isEmpty()) {
        return;
    }

    line->setText(fn);
}

int UIPreferences::getKeyInput() {
    UIKeyInputDialog d(window());

    if(d.exec()) {
        return d.getKeyPressed();
    }

    return 0;
}

void UIPreferences::setKey(const char *buttonmap, const char *button,
                           int console) {
    QVariantMap map;
    QVariantMap *bm = &sms_bm;

    if(console == CONSOLE_COLECOVISION)
        bm = &coleco_bm;
    else if(console == CONSOLE_NES)
        bm = &nes_bm;

    int key = getKeyInput();

    if(!key)
        return;

    map = bm->value(buttonmap).toMap();

    map[button] = key;
    (*bm)[buttonmap] = map;
}
