/*
    This file is part of CrabEmu.

    Copyright (C) 2005, 2006, 2007, 2008, 2009, 2012, 2013 Lawrence Sebald

    CrabEmu is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2
    as published by the Free Software Foundation.

    CrabEmu is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CrabEmu; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <QApplication>
#include <QMutex>
#include <QMessageBox>

#include <cstring>

#include "gui/UICrabEmu.h"
#include "gui/UIDebugWindow.h"
#include "EmuWidget.h"

#include "sms.h"
#include "colecovision.h"
#include "nes.h"
#include "smsvdp.h"
#include "nesppu.h"

static bool resized = false;
extern "C" sms_vdp_t smsvdp;
extern "C" int sms_initialized;
extern "C" int sms_console;
static UICrabEmu *mainWindow;

QMutex memoryLock;

using std::memcpy;

void render_scene(void) {
#ifndef ENABLE_OPENGL
    QImage *img = mainWindow->emuwidget->retainImage();
    uchar *bits = img->bits();

    if(sms_console != CONSOLE_NES) {
        /* Simple, but effective. */
#ifdef CRABEMU_32BIT_COLOR
        memcpy(bits, smsvdp.framebuffer, 256 * 256 * 4);
#else
        memcpy(bits, smsvdp.framebuffer, 256 * 256 * 2);
#endif
    }
    else {
#ifdef CRABEMU_32BIT_COLOR
        memcpy(bits, nes_ppu_framebuffer(), 256 * 256 * 4);
#else
        memcpy(bits, nes_ppu_framebuffer(), 256 * 256 * 2);
#endif
    }

    mainWindow->emuwidget->releaseImage();
#endif
}

int emu_frame(int run) {
    switch(sms_console) {
        case CONSOLE_COLECOVISION:
            run = coleco_frame(run, 0);
            break;

        case CONSOLE_NES:
            run = nes_frame(run, 0);
            break;

        case CONSOLE_SMS:
        case CONSOLE_GG:
        case CONSOLE_SG1000:
        case CONSOLE_SC3000:
            run = sms_frame(run, 0);
            break;
    }

    return run;
}

extern "C" void gui_set_aspect(float x __UNUSED__, float y __UNUSED__) {
    /* XXXX: Do something about this function sometime. */
    resized = true;
}

extern "C" void gui_set_title(const char *str) {
    mainWindow->setWindowTitle(QString::fromUtf8(str));
}

#ifdef ENABLE_SDSC_TERMINAL

extern "C" void gui_update_sdsc_terminal(char console[25][80 * 2]) {
    UIDebugWindow *d = mainWindow->debug();
    if(d) {
        d->updateSDSC(console);
    }
}

#endif

int main(int argc, char *argv[]) {
    QApplication app(argc, argv);
    QStringList args;

    QApplication::setOrganizationName("ljsdcdev");
    QApplication::setOrganizationDomain("ljsdcdev.net");
    QApplication::setApplicationName("CrabEmu");

    mainWindow = new UICrabEmu;
    mainWindow->setWindowTitle(app.applicationName());
    mainWindow->restoreSettings();
    mainWindow->show();

    QObject::connect(&app, SIGNAL(lastWindowClosed()), &app, SLOT(quit()));

    args = app.arguments();
    if (!args.empty() && args.count() > 1) {
        mainWindow->loadRom(args.at(1));
    }

    return app.exec();
}
